# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ply.lex as lex
import six

operators = {
    'age': 'OP_AGE',
    'recentlyseen': 'OP_RECENTLYSEEN', # Gertty extension
    'change': 'OP_CHANGE',
    'owner': 'OP_OWNER',
    #'OP_OWNERIN', # needs local group membership
    'reviewer': 'OP_REVIEWER',
    #'OP_REVIEWERIN', # needs local group membership
    'commit': 'OP_COMMIT',
    'project': 'OP_PROJECT',
    'projects': 'OP_PROJECTS',
    '_project_key': 'OP_PROJECT_KEY',  # internal gertty use only
    'branch': 'OP_BRANCH',
    'topic': 'OP_TOPIC',
    'ref': 'OP_REF',
    #'tr': 'OP_TR', # needs trackingids
    #'bug': 'OP_BUG', # needs trackingids
    'label': 'OP_LABEL',
    'message': 'OP_MESSAGE',
    'comment': 'OP_COMMENT',
    'file': 'OP_FILE',
    'has': 'OP_HAS',
    'is': 'OP_IS',
    'status': 'OP_STATUS',
    'limit': 'OP_LIMIT',
    }

reserved = {
    'or|OR': 'OR',
    'not|NOT': 'NOT',
    }

tokens = [
    'OP',
    'AND',
    'OR',
    'NOT',
    'NEG',
    'LPAREN',
    'RPAREN',
    'NUMBER',
    'CHANGE_ID',
    'SSTRING',
    'DSTRING',
    'USTRING',
    #'REGEX',
    #'SHA',
    ] + list(operators.values())

def SearchTokenizer():
    t_LPAREN = r'\('   # NOQA
    t_RPAREN = r'\)'   # NOQA
    t_NEG    = r'[-!]' # NOQA
    t_ignore = ' \t'   # NOQA (and intentionally not using r'' due to tab char)

    def t_OP(t):
        r'[a-zA-Z_][a-zA-Z_]*:'
        t.type = operators.get(t.value[:-1], 'OP')
        return t

    def t_CHANGE_ID(t):
        r'I[a-fA-F0-9]{7,40}'
        return t

    def t_SSTRING(t):
        r"'([^\\']+|\\'|\\\\)*'"
        t.value=t.value[1:-1].decode("string-escape")
        return t

    def t_DSTRING(t):
        r'"([^\\"]+|\\"|\\\\)*"'
        t.value=t.value[1:-1].decode("string-escape")
        return t

    def t_AND(t):
        r'and|AND'
        return t

    def t_OR(t):
        r'or|OR'
        return t

    def t_NOT(t):
        r'not|NOT'
        return t

    def t_INTEGER(t):
        r'[+-]\d+'
        t.value = int(t.value)
        return t

    def t_NUMBER(t):
        r'\d+'
        t.value = int(t.value)
        return t

    def t_USTRING(t):
        r'([^\s\(\)!-][^\s\(\)!]*)'
        t.value = six.b(t.value).decode("unicode_escape")
        return t

    def t_newline(t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    def t_error(t):
        print("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    return lex.lex()



# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urwid

from gertty import keymap
from gertty import mywid
from gertty.view.diff import BaseDiffComment, BaseDiffCommentEdit, BaseDiffLine
from gertty.view.diff import BaseFileHeader, BaseFileReminder, BaseDiffView

LN_COL_WIDTH = 5

class SideDiffCommentEdit(BaseDiffCommentEdit):
    def __init__(self, app, context, old_key=None, new_key=None, old=u'', new=u''):
        super(SideDiffCommentEdit, self).__init__([])
        self.app = app
        self.context = context
        # If we save a comment, the resulting key will be stored here
        self.old_key = old_key
        self.new_key = new_key
        self.old = mywid.MyEdit(edit_text=old, multiline=True, ring=app.ring)
        self.new = mywid.MyEdit(edit_text=new, multiline=True, ring=app.ring)
        self.contents.append((urwid.Text(u''), ('given', LN_COL_WIDTH, False)))
        if context.old_file_key and (context.old_ln is not None or context.header):
            self.contents.append((urwid.AttrMap(self.old, 'draft-comment'), ('weight', 1, False)))
        else:
            self.contents.append((urwid.Text(u''), ('weight', 1, False)))
        self.contents.append((urwid.Text(u''), ('given', LN_COL_WIDTH, False)))
        if context.new_file_key and (context.new_ln is not None or context.header):
            self.contents.append((urwid.AttrMap(self.new, 'draft-comment'), ('weight', 1, False)))
            new_editable = True
        else:
            self.contents.append((urwid.Text(u''), ('weight', 1, False)))
            new_editable = False
        if new_editable:
            self.focus_position = 3
        else:
            self.focus_position = 1

    def keypress(self, size, key):
        if not self.app.input_buffer:
            key = super(SideDiffCommentEdit, self).keypress(size, key)
        keys = self.app.input_buffer + [key]
        commands = self.app.config.keymap.getCommands(keys)

        if ((keymap.NEXT_SELECTABLE in commands) or
            (keymap.PREV_SELECTABLE in commands)):
            if ((self.context.old_ln is not None and
                 self.context.new_ln is not None) or
                self.context.header):
                if self.focus_position == 3:
                    self.focus_position = 1
                else:
                    self.focus_position = 3
            return None
        return key

class SideDiffComment(BaseDiffComment):
    def __init__(self, context, old, new):
        super(SideDiffComment, self).__init__([])
        self.context = context
        oldt = urwid.Text(old)
        newt = urwid.Text(new)
        if old:
            oldt = urwid.AttrMap(oldt, 'comment')
        if new:
            newt = urwid.AttrMap(newt, 'comment')
        self.contents.append((urwid.Text(u''), ('given', LN_COL_WIDTH, False)))
        self.contents.append((oldt, ('weight', 1, False)))
        self.contents.append((urwid.Text(u''), ('given', LN_COL_WIDTH, False)))
        self.contents.append((newt, ('weight', 1, False)))

class SideDiffLine(BaseDiffLine):
    def __init__(self, app, context, old, new, callback=None):
        super(SideDiffLine, self).__init__('', on_press=callback)
        self.context = context
        self.text_widgets = []
        columns = []
        for (ln, action, line) in (old, new):
            if ln is None:
                ln = ''
            else:
                ln = '%*i' % (LN_COL_WIDTH-1, ln)
            ln_col = urwid.Text(('line-number', ln))
            ln_col.set_wrap_mode('clip')
            line_col = mywid.SearchableText(line)
            self.text_widgets.append(line_col)

            if action == '':
                line_col = urwid.AttrMap(line_col, 'nonexistent')
            columns += [(LN_COL_WIDTH, ln_col), line_col]
        col = urwid.Columns(columns)
        map = {None: 'focused',
               'added-line': 'focused-added-line',
               'added-word': 'focused-added-word',
               'removed-line': 'focused-removed-line',
               'removed-word': 'focused-removed-word',
               'nonexistent': 'focused-nonexistent',
               'line-number': 'focused-line-number',
               }
        self._w = urwid.AttrMap(col, None, focus_map=map)

    def search(self, search, attribute):
        ret = False
        for w in self.text_widgets:
            if w.search(search, attribute):
                ret = True
        return ret

class SideFileHeader(BaseFileHeader):
    def __init__(self, app, context, old, new, callback=None):
        super(SideFileHeader, self).__init__('', on_press=callback)
        self.context = context
        col = urwid.Columns([
                urwid.Text(('filename', old)),
                urwid.Text(('filename', new))])
        map = {None: 'focused-filename',
               'filename': 'focused-filename'}
        self._w = urwid.AttrMap(col, None, focus_map=map)

class SideFileReminder(BaseFileReminder):
    def __init__(self):
        self.old_text = urwid.Text(('filename', ''))
        self.new_text = urwid.Text(('filename', ''))
        col = urwid.Columns([self.old_text, self.new_text])
        super(SideFileReminder, self).__init__(col)

    def set(self, old, new):
        self.old_text.set_text(('filename', old))
        self.new_text.set_text(('filename', new))

class SideDiffView(BaseDiffView):
    def makeLines(self, diff, lines_to_add, comment_lists):
        lines = []
        for old, new in lines_to_add:
            context = self.makeContext(diff, old[0], new[0])
            lines.append(SideDiffLine(self.app, context, old, new,
                                      callback=self.onSelect))
            # see if there are any comments for this line
            key = 'old-%s-%s' % (old[0], diff.oldname)
            old_list = comment_lists.pop(key, [])
            key = 'new-%s-%s' % (new[0], diff.newname)
            new_list = comment_lists.pop(key, [])
            while old_list or new_list:
                old_comment_key = new_comment_key = None
                old_comment = new_comment = u''
                if old_list:
                    (old_comment_key, old_comment) = old_list.pop(0)
                if new_list:
                    (new_comment_key, new_comment) = new_list.pop(0)
                lines.append(SideDiffComment(context, old_comment, new_comment))
            # see if there are any draft comments for this line
            key = 'olddraft-%s-%s' % (old[0], diff.oldname)
            old_list = comment_lists.pop(key, [])
            key = 'newdraft-%s-%s' % (new[0], diff.newname)
            new_list = comment_lists.pop(key, [])
            while old_list or new_list:
                old_comment_key = new_comment_key = None
                old_comment = new_comment = u''
                if old_list:
                    (old_comment_key, old_comment) = old_list.pop(0)
                if new_list:
                    (new_comment_key, new_comment) = new_list.pop(0)
                lines.append(SideDiffCommentEdit(self.app, context,
                                                 old_comment_key,
                                                 new_comment_key,
                                                 old_comment, new_comment))
        return lines

    def makeFileReminder(self):
        return SideFileReminder()

    def makeFileHeader(self, diff, comment_lists):
        context = self.makeContext(diff, None, None, header=True)
        lines = []
        lines.append(SideFileHeader(self.app, context, diff.oldname, diff.newname,
                                    callback=self.onSelect))

        # see if there are any comments for this file
        key = 'old-None-%s' % (diff.oldname,)
        old_list = comment_lists.pop(key, [])
        key = 'new-None-%s' % (diff.newname,)
        new_list = comment_lists.pop(key, [])
        while old_list or new_list:
            old_comment_key = new_comment_key = None
            old_comment = new_comment = u''
            if old_list:
                (old_comment_key, old_comment) = old_list.pop(0)
            if new_list:
                (new_comment_key, new_comment) = new_list.pop(0)
            lines.append(SideDiffComment(context, old_comment, new_comment))
        # see if there are any draft comments for this file
        key = 'olddraft-None-%s' % (diff.oldname,)
        old_list = comment_lists.pop(key, [])
        key = 'newdraft-None-%s' % (diff.newname,)
        new_list = comment_lists.pop(key, [])
        while old_list or new_list:
            old_comment_key = new_comment_key = None
            old_comment = new_comment = u''
            if old_list:
                (old_comment_key, old_comment) = old_list.pop(0)
            if new_list:
                (new_comment_key, new_comment) = new_list.pop(0)
            lines.append(SideDiffCommentEdit(self.app, context,
                                             old_comment_key,
                                             new_comment_key,
                                             old_comment, new_comment))
        return lines

    def makeCommentEdit(self, edit):
        return SideDiffCommentEdit(self.app, edit.context)

    def cleanupEdit(self, edit):
        if edit.old_key:
            self.deleteComment(edit.old_key)
            edit.old_key = None
        if edit.new_key:
            self.deleteComment(edit.new_key)
            edit.new_key = None
        old = edit.old.edit_text.strip()
        new = edit.new.edit_text.strip()
        if old or new:
            if old:
                edit.old_key = self.saveComment(
                    edit.context, old, new=False)
            if new:
                edit.new_key = self.saveComment(
                    edit.context, new, new=True)
        else:
            self.listbox.body.remove(edit)



# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
try:
    import ordereddict
except:
    pass
import re

import six
import urwid

from gertty import mywid

try:
    OrderedDict = collections.OrderedDict
except AttributeError:
    OrderedDict = ordereddict.OrderedDict

class TextReplacement(object):
    def __init__(self, config):
        if isinstance(config, six.string_types):
            self.color = None
            self.text = config
        else:
            self.color = config.get('color')
            self.text = config['text']

    def replace(self, app, data):
        if self.color:
            return (self.color.format(**data), self.text.format(**data))
        return (None, self.text.format(**data))

class LinkReplacement(object):
    def __init__(self, config):
        self.url = config['url']
        self.text = config['text']

    def replace(self, app, data):
        link = mywid.Link(self.text.format(**data), 'link', 'focused-link')
        urwid.connect_signal(link, 'selected',
            lambda link:self.activate(app, self.url.format(**data)))
        return link

    def activate(self, app, url):
        result = app.parseInternalURL(url)
        if result is not None:
            return app.openInternalURL(result)
        return app.openURL(url)

class SearchReplacement(object):
    def __init__(self, config):
        self.query = config['query']
        self.text = config['text']

    def replace(self, app, data):
        link = mywid.Link(self.text.format(**data), 'link', 'focused-link')
        urwid.connect_signal(link, 'selected',
            lambda link:app.doSearch(self.query.format(**data)))
        return link

class CommentLink(object):
    def __init__(self, config):
        self.match = re.compile(config['match'], re.M)
        self.test_result = config.get('test-result', None)
        self.replacements = []
        for r in config['replacements']:
            if 'text' in r:
                self.replacements.append(TextReplacement(r['text']))
            if 'link' in r:
                self.replacements.append(LinkReplacement(r['link']))
            if 'search' in r:
                self.replacements.append(SearchReplacement(r['search']))

    def getTestResults(self, app, text):
        if self.test_result is None:
            return {}
        ret = OrderedDict()
        for line in text.split('\n'):
            m = self.match.search(line)
            if m:
                repl = [r.replace(app, m.groupdict()) for r in self.replacements]
                job = self.test_result.format(**m.groupdict())
                ret[job] = repl + ['\n']
        return ret

    def run(self, app, chunks):
        ret = []
        for chunk in chunks:
            if not isinstance(chunk, six.string_types):
                # We don't currently support nested commentlinks; if
                # we have something that isn't a string, just append
                # it to the output.
                ret.append(chunk)
                continue
            if not chunk:
                ret += [chunk]
            while chunk:
                m = self.match.search(chunk)
                if not m:
                    ret.append(chunk)
                    break
                before = chunk[:m.start()]
                after = chunk[m.end():]
                if before:
                    ret.append(before)
                ret += [r.replace(app, m.groupdict()) for r in self.replacements]
                chunk = after
        return ret




"""add topic table

Revision ID: 4388de50824a
Revises: 254ac5fc3941
Create Date: 2015-10-31 19:06:38.538948

"""

# revision identifiers, used by Alembic.
revision = '4388de50824a'
down_revision = '254ac5fc3941'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('topic',
    sa.Column('key', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=255), index=True, nullable=False),
    sa.Column('sequence', sa.Integer(), index=True, unique=True, nullable=False),
    sa.PrimaryKeyConstraint('key')
    )

    op.create_table('project_topic',
    sa.Column('key', sa.Integer(), nullable=False),
    sa.Column('project_key', sa.Integer(), sa.ForeignKey('project.key'), index=True),
    sa.Column('topic_key', sa.Integer(), sa.ForeignKey('topic.key'), index=True),
    sa.Column('sequence', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('key'),
    sa.UniqueConstraint('topic_key', 'sequence', name='topic_key_sequence_const'),
    )

def downgrade():
    pass



"""Initial schema

Revision ID: 44402069e137
Revises: None
Create Date: 2014-05-04 17:10:23.127702

"""

# revision identifiers, used by Alembic.
revision = '44402069e137'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('project',
    sa.Column('key', sa.Integer(), nullable=False, quote=True),
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.Column('subscribed', sa.Boolean(), nullable=True),
    sa.Column('description', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_index(op.f('ix_project_name'), 'project', ['name'], unique=True)
    op.create_index(op.f('ix_project_subscribed'), 'project', ['subscribed'], unique=False)
    op.create_table('change',
    sa.Column('key', sa.Integer(), nullable=False, quote=True),
    sa.Column('project_key', sa.Integer(), nullable=True),
    sa.Column('id', sa.String(length=255), nullable=False),
    sa.Column('number', sa.Integer(), nullable=False),
    sa.Column('branch', sa.String(length=255), nullable=False),
    sa.Column('change_id', sa.String(length=255), nullable=False),
    sa.Column('topic', sa.String(length=255), nullable=True),
    sa.Column('owner', sa.String(length=255), nullable=True),
    sa.Column('subject', sa.Text(), nullable=False),
    sa.Column('created', sa.DateTime(), nullable=False),
    sa.Column('updated', sa.DateTime(), nullable=False),
    sa.Column('status', sa.String(length=8), nullable=False),
    sa.Column('hidden', sa.Boolean(), nullable=False),
    sa.Column('reviewed', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['project_key'], ['project.key'], ),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_index(op.f('ix_change_branch'), 'change', ['branch'], unique=False)
    op.create_index(op.f('ix_change_change_id'), 'change', ['change_id'], unique=False)
    op.create_index(op.f('ix_change_created'), 'change', ['created'], unique=False)
    op.create_index(op.f('ix_change_hidden'), 'change', ['hidden'], unique=False)
    op.create_index(op.f('ix_change_id'), 'change', ['id'], unique=True)
    op.create_index(op.f('ix_change_number'), 'change', ['number'], unique=True)
    op.create_index(op.f('ix_change_owner'), 'change', ['owner'], unique=False)
    op.create_index(op.f('ix_change_project_key'), 'change', ['project_key'], unique=False)
    op.create_index(op.f('ix_change_reviewed'), 'change', ['reviewed'], unique=False)
    op.create_index(op.f('ix_change_status'), 'change', ['status'], unique=False)
    op.create_index(op.f('ix_change_topic'), 'change', ['topic'], unique=False)
    op.create_index(op.f('ix_change_updated'), 'change', ['updated'], unique=False)
    op.create_table('approval',
    sa.Column('key', sa.Integer(), nullable=False, quote=True),
    sa.Column('change_key', sa.Integer(), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('category', sa.String(length=255), nullable=False),
    sa.Column('value', sa.Integer(), nullable=False),
    sa.Column('pending', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['change_key'], ['change.key'], ),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_index(op.f('ix_approval_change_key'), 'approval', ['change_key'], unique=False)
    op.create_index(op.f('ix_approval_pending'), 'approval', ['pending'], unique=False)
    op.create_table('revision',
    sa.Column('key', sa.Integer(), nullable=False, quote=True),
    sa.Column('change_key', sa.Integer(), nullable=True),
    sa.Column('number', sa.Integer(), nullable=False),
    sa.Column('message', sa.Text(), nullable=False),
    sa.Column('commit', sa.String(length=255), nullable=False),
    sa.Column('parent', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['change_key'], ['change.key'], ),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_index(op.f('ix_revision_change_key'), 'revision', ['change_key'], unique=False)
    op.create_index(op.f('ix_revision_number'), 'revision', ['number'], unique=False)
    op.create_table('label',
    sa.Column('key', sa.Integer(), nullable=False, quote=True),
    sa.Column('change_key', sa.Integer(), nullable=True),
    sa.Column('category', sa.String(length=255), nullable=False),
    sa.Column('value', sa.Integer(), nullable=False),
    sa.Column('description', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['change_key'], ['change.key'], ),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_index(op.f('ix_label_change_key'), 'label', ['change_key'], unique=False)
    op.create_table('permitted_label',
    sa.Column('key', sa.Integer(), nullable=False, quote=True),
    sa.Column('change_key', sa.Integer(), nullable=True),
    sa.Column('category', sa.String(length=255), nullable=False),
    sa.Column('value', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['change_key'], ['change.key'], ),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_index(op.f('ix_permitted_label_change_key'), 'permitted_label', ['change_key'], unique=False)
    op.create_table('comment',
    sa.Column('key', sa.Integer(), nullable=False, quote=True),
    sa.Column('revision_key', sa.Integer(), nullable=True),
    sa.Column('id', sa.String(length=255), nullable=True),
    sa.Column('in_reply_to', sa.String(length=255), nullable=True),
    sa.Column('created', sa.DateTime(), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('file', sa.Text(), nullable=False),
    sa.Column('parent', sa.Boolean(), nullable=False),
    sa.Column('line', sa.Integer(), nullable=True),
    sa.Column('message', sa.Text(), nullable=False),
    sa.Column('pending', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['revision_key'], ['revision.key'], ),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_index(op.f('ix_comment_created'), 'comment', ['created'], unique=False)
    op.create_index(op.f('ix_comment_id'), 'comment', ['id'], unique=False)
    op.create_index(op.f('ix_comment_pending'), 'comment', ['pending'], unique=False)
    op.create_index(op.f('ix_comment_revision_key'), 'comment', ['revision_key'], unique=False)
    op.create_table('message',
    sa.Column('key', sa.Integer(), nullable=False, quote=True),
    sa.Column('revision_key', sa.Integer(), nullable=True),
    sa.Column('id', sa.String(length=255), nullable=True),
    sa.Column('created', sa.DateTime(), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('message', sa.Text(), nullable=False),
    sa.Column('pending', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['revision_key'], ['revision.key'], ),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_index(op.f('ix_message_created'), 'message', ['created'], unique=False)
    op.create_index(op.f('ix_message_id'), 'message', ['id'], unique=False)
    op.create_index(op.f('ix_message_pending'), 'message', ['pending'], unique=False)
    op.create_index(op.f('ix_message_revision_key'), 'message', ['revision_key'], unique=False)
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_message_revision_key'), table_name='message')
    op.drop_index(op.f('ix_message_pending'), table_name='message')
    op.drop_index(op.f('ix_message_id'), table_name='message')
    op.drop_index(op.f('ix_message_created'), table_name='message')
    op.drop_table('message')
    op.drop_index(op.f('ix_comment_revision_key'), table_name='comment')
    op.drop_index(op.f('ix_comment_pending'), table_name='comment')
    op.drop_index(op.f('ix_comment_id'), table_name='comment')
    op.drop_index(op.f('ix_comment_created'), table_name='comment')
    op.drop_table('comment')
    op.drop_index(op.f('ix_permitted_label_change_key'), table_name='permitted_label')
    op.drop_table('permitted_label')
    op.drop_index(op.f('ix_label_change_key'), table_name='label')
    op.drop_table('label')
    op.drop_index(op.f('ix_revision_number'), table_name='revision')
    op.drop_index(op.f('ix_revision_change_key'), table_name='revision')
    op.drop_table('revision')
    op.drop_index(op.f('ix_approval_pending'), table_name='approval')
    op.drop_index(op.f('ix_approval_change_key'), table_name='approval')
    op.drop_table('approval')
    op.drop_index(op.f('ix_change_updated'), table_name='change')
    op.drop_index(op.f('ix_change_topic'), table_name='change')
    op.drop_index(op.f('ix_change_status'), table_name='change')
    op.drop_index(op.f('ix_change_reviewed'), table_name='change')
    op.drop_index(op.f('ix_change_project_key'), table_name='change')
    op.drop_index(op.f('ix_change_owner'), table_name='change')
    op.drop_index(op.f('ix_change_number'), table_name='change')
    op.drop_index(op.f('ix_change_id'), table_name='change')
    op.drop_index(op.f('ix_change_hidden'), table_name='change')
    op.drop_index(op.f('ix_change_created'), table_name='change')
    op.drop_index(op.f('ix_change_change_id'), table_name='change')
    op.drop_index(op.f('ix_change_branch'), table_name='change')
    op.drop_table('change')
    op.drop_index(op.f('ix_project_subscribed'), table_name='project')
    op.drop_index(op.f('ix_project_name'), table_name='project')
    op.drop_table('project')
    ### end Alembic commands ###



# Copyright 2014 Mirantis Inc.
# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six
import uuid

from alembic import op
import sqlalchemy


def sqlite_alter_columns(table_name, column_defs):
    """Implement alter columns for SQLite.

    The ALTER COLUMN command isn't supported by SQLite specification.
    Instead of calling ALTER COLUMN it uses the following workaround:

    * create temp table '{table_name}_{rand_uuid}', with some column
      defs replaced;
    * copy all data to the temp table;
    * drop old table;
    * rename temp table to the old table name.
    """
    connection = op.get_bind()
    meta = sqlalchemy.MetaData(bind=connection)
    meta.reflect()

    changed_columns = {}
    indexes = []
    for col in column_defs:
        # If we are to have an index on the column, don't create it
        # immediately, instead, add it to a list of indexes to create
        # after the table rename.
        if col.index:
            indexes.append(('ix_%s_%s' % (table_name, col.name),
                            table_name,
                            [col.name],
                            col.unique))
            col.unique = False
            col.index = False
        changed_columns[col.name] = col

    # construct lists of all columns and their names
    old_columns = []
    new_columns = []
    column_names = []
    for column in meta.tables[table_name].columns:
        column_names.append(column.name)
        old_columns.append(column)
        if column.name in changed_columns.keys():
            new_columns.append(changed_columns[column.name])
        else:
            col_copy = column.copy()
            new_columns.append(col_copy)

    for key in meta.tables[table_name].foreign_keys:
        constraint = key.constraint
        con_copy = constraint.copy()
        new_columns.append(con_copy)

    for index in meta.tables[table_name].indexes:
        # If this is a single column index for a changed column, don't
        # copy it because we may already be creating a new version of
        # it (or removing it).
        idx_columns = [col.name for col in index.columns]
        if len(idx_columns)==1 and idx_columns[0] in changed_columns.keys():
            continue
        # Otherwise, recreate the index.
        indexes.append((index.name,
                        table_name,
                        [col.name for col in index.columns],
                        index.unique))

    # create temp table
    tmp_table_name = "%s_%s" % (table_name, six.text_type(uuid.uuid4()))
    op.create_table(tmp_table_name, *new_columns)
    meta.reflect()

    try:
        # copy data from the old table to the temp one
        sql_select = sqlalchemy.sql.select(old_columns)
        connection.execute(sqlalchemy.sql.insert(meta.tables[tmp_table_name])
                           .from_select(column_names, sql_select))
    except Exception:
        op.drop_table(tmp_table_name)
        raise

    # drop the old table and rename temp table to the old table name
    op.drop_table(table_name)
    op.rename_table(tmp_table_name, table_name)

    # (re-)create indexes
    for index in indexes:
        op.create_index(op.f(index[0]), index[1], index[2], unique=index[3])

def sqlite_drop_columns(table_name, drop_columns):
    """Implement drop columns for SQLite.

    The DROP COLUMN command isn't supported by SQLite specification.
    Instead of calling DROP COLUMN it uses the following workaround:

    * create temp table '{table_name}_{rand_uuid}', without
      dropped columns;
    * copy all data to the temp table;
    * drop old table;
    * rename temp table to the old table name.
    """
    connection = op.get_bind()
    meta = sqlalchemy.MetaData(bind=connection)
    meta.reflect()

    # construct lists of all columns and their names
    old_columns = []
    new_columns = []
    column_names = []
    indexes = []
    for column in meta.tables[table_name].columns:
        if column.name not in drop_columns:
            old_columns.append(column)
            column_names.append(column.name)
            col_copy = column.copy()
            new_columns.append(col_copy)

    for key in meta.tables[table_name].foreign_keys:
        # If this is a single column constraint for a dropped column,
        # don't copy it.
        if isinstance(key.constraint.columns, sqlalchemy.sql.base.ColumnCollection):
            # This is needed for SQLAlchemy >= 1.0.4
            columns = [c.name for c in key.constraint.columns]
        else:
            # This is needed for SQLAlchemy <= 0.9.9.  This is
            # backwards compat code just in case someone updates
            # Gertty without updating SQLAlchemy.  This is simple
            # enough to check and will hopefully avoid leaving the
            # user's db in an inconsistent state.  Remove this after
            # Gertty 1.2.0.
            columns = key.constraint.columns
        if (len(columns)==1 and columns[0] in drop_columns):
            continue
        # Otherwise, recreate the constraint.
        constraint = key.constraint
        con_copy = constraint.copy()
        new_columns.append(con_copy)

    for index in meta.tables[table_name].indexes:
        # If this is a single column index for a dropped column, don't
        # copy it.
        idx_columns = [col.name for col in index.columns]
        if len(idx_columns)==1 and idx_columns[0] in drop_columns:
            continue
        # Otherwise, recreate the index.
        indexes.append((index.name,
                        table_name,
                        [col.name for col in index.columns],
                        index.unique))

    # create temp table
    tmp_table_name = "%s_%s" % (table_name, six.text_type(uuid.uuid4()))
    op.create_table(tmp_table_name, *new_columns)
    meta.reflect()

    try:
        # copy data from the old table to the temp one
        sql_select = sqlalchemy.sql.select(old_columns)
        connection.execute(sqlalchemy.sql.insert(meta.tables[tmp_table_name])
                           .from_select(column_names, sql_select))
    except Exception:
        op.drop_table(tmp_table_name)
        raise

    # drop the old table and rename temp table to the old table name
    op.drop_table(table_name)
    op.rename_table(tmp_table_name, table_name)

    # (re-)create indexes
    for index in indexes:
        op.create_index(op.f(index[0]), index[1], index[2], unique=index[3])



"""add last_seen column to change

Revision ID: 37a702b7f58e
Revises: 3610c2543e07
Create Date: 2016-02-06 09:09:38.728225

"""

# revision identifiers, used by Alembic.
revision = '37a702b7f58e'
down_revision = '3610c2543e07'

import warnings

from alembic import op
import sqlalchemy as sa


def upgrade():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        op.add_column('change', sa.Column('last_seen', sa.DateTime, index=True))


def downgrade():
    pass



# Copyright 2014 OpenStack Foundation
# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import time
import logging
import threading

import alembic
import alembic.config
import six
import sqlalchemy
from sqlalchemy import create_engine, MetaData, Table, Column, Integer, String, Boolean, DateTime, Text, UniqueConstraint
from sqlalchemy.schema import ForeignKey
from sqlalchemy.orm import mapper, sessionmaker, relationship, scoped_session, joinedload
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import exists
from sqlalchemy.sql.expression import and_

metadata = MetaData()
project_table = Table(
    'project', metadata,
    Column('key', Integer, primary_key=True),
    Column('name', String(255), index=True, unique=True, nullable=False),
    Column('subscribed', Boolean, index=True, default=False),
    Column('description', Text, nullable=False, default=''),
    Column('updated', DateTime, index=True),
    )
branch_table = Table(
    'branch', metadata,
    Column('key', Integer, primary_key=True),
    Column('project_key', Integer, ForeignKey("project.key"), index=True),
    Column('name', String(255), index=True, nullable=False),
    )
topic_table = Table(
    'topic', metadata,
    Column('key', Integer, primary_key=True),
    Column('name', String(255), index=True, nullable=False),
    Column('sequence', Integer, index=True, unique=True, nullable=False),
    )
project_topic_table = Table(
    'project_topic', metadata,
    Column('key', Integer, primary_key=True),
    Column('project_key', Integer, ForeignKey("project.key"), index=True),
    Column('topic_key', Integer, ForeignKey("topic.key"), index=True),
    Column('sequence', Integer, nullable=False),
    UniqueConstraint('topic_key', 'sequence', name='topic_key_sequence_const'),
    )
change_table = Table(
    'change', metadata,
    Column('key', Integer, primary_key=True),
    Column('project_key', Integer, ForeignKey("project.key"), index=True),
    Column('id', String(255), index=True, unique=True, nullable=False),
    Column('number', Integer, index=True, unique=True, nullable=False),
    Column('branch', String(255), index=True, nullable=False),
    Column('change_id', String(255), index=True, nullable=False),
    Column('topic', String(255), index=True),
    Column('account_key', Integer, ForeignKey("account.key"), index=True),
    Column('subject', Text, nullable=False),
    Column('created', DateTime, index=True, nullable=False),
    Column('updated', DateTime, index=True, nullable=False),
    Column('status', String(16), index=True, nullable=False),
    Column('hidden', Boolean, index=True, nullable=False),
    Column('reviewed', Boolean, index=True, nullable=False),
    Column('starred', Boolean, index=True, nullable=False),
    Column('held', Boolean, index=True, nullable=False),
    Column('pending_rebase', Boolean, index=True, nullable=False),
    Column('pending_topic', Boolean, index=True, nullable=False),
    Column('pending_starred', Boolean, index=True, nullable=False),
    Column('pending_status', Boolean, index=True, nullable=False),
    Column('pending_status_message', Text),
    Column('last_seen', DateTime, index=True),
    Column('outdated', Boolean, index=True, nullable=False),
    )
change_conflict_table = Table(
    'change_conflict', metadata,
    Column('key', Integer, primary_key=True),
    Column('change1_key', Integer, ForeignKey("change.key"), index=True),
    Column('change2_key', Integer, ForeignKey("change.key"), index=True),
    )
revision_table = Table(
    'revision', metadata,
    Column('key', Integer, primary_key=True),
    Column('change_key', Integer, ForeignKey("change.key"), index=True),
    Column('number', Integer, index=True, nullable=False),
    Column('message', Text, nullable=False),
    Column('commit', String(255), index=True, nullable=False),
    Column('parent', String(255), index=True, nullable=False),
    # TODO: fetch_ref, fetch_auth are unused; remove
    Column('fetch_auth', Boolean, nullable=False),
    Column('fetch_ref', String(255), nullable=False),
    Column('pending_message', Boolean, index=True, nullable=False),
    Column('can_submit', Boolean, nullable=False),
    )
message_table = Table(
    'message', metadata,
    Column('key', Integer, primary_key=True),
    Column('revision_key', Integer, ForeignKey("revision.key"), index=True),
    Column('account_key', Integer, ForeignKey("account.key"), index=True),
    Column('id', String(255), index=True), #, unique=True, nullable=False),
    Column('created', DateTime, index=True, nullable=False),
    Column('message', Text, nullable=False),
    Column('draft', Boolean, index=True, nullable=False),
    Column('pending', Boolean, index=True, nullable=False),
    )
comment_table = Table(
    'comment', metadata,
    Column('key', Integer, primary_key=True),
    Column('file_key', Integer, ForeignKey("file.key"), index=True),
    Column('account_key', Integer, ForeignKey("account.key"), index=True),
    Column('id', String(255), index=True), #, unique=True, nullable=False),
    Column('in_reply_to', String(255)),
    Column('created', DateTime, index=True, nullable=False),
    Column('parent', Boolean, nullable=False),
    Column('line', Integer),
    Column('message', Text, nullable=False),
    Column('draft', Boolean, index=True, nullable=False),
    )
label_table = Table(
    'label', metadata,
    Column('key', Integer, primary_key=True),
    Column('change_key', Integer, ForeignKey("change.key"), index=True),
    Column('category', String(255), nullable=False),
    Column('value', Integer, nullable=False),
    Column('description', String(255), nullable=False),
    )
permitted_label_table = Table(
    'permitted_label', metadata,
    Column('key', Integer, primary_key=True),
    Column('change_key', Integer, ForeignKey("change.key"), index=True),
    Column('category', String(255), nullable=False),
    Column('value', Integer, nullable=False),
    )
approval_table = Table(
    'approval', metadata,
    Column('key', Integer, primary_key=True),
    Column('change_key', Integer, ForeignKey("change.key"), index=True),
    Column('account_key', Integer, ForeignKey("account.key"), index=True),
    Column('category', String(255), nullable=False),
    Column('value', Integer, nullable=False),
    Column('draft', Boolean, index=True, nullable=False),
    )
account_table = Table(
    'account', metadata,
    Column('key', Integer, primary_key=True),
    Column('id', Integer, index=True, unique=True, nullable=False),
    Column('name', String(255), index=True),
    Column('username', String(255), index=True),
    Column('email', String(255), index=True),
    )
pending_cherry_pick_table = Table(
    'pending_cherry_pick', metadata,
    Column('key', Integer, primary_key=True),
    Column('revision_key', Integer, ForeignKey("revision.key"), index=True),
    # Branch is a str here to avoid FK complications if the branch
    # entry is removed.
    Column('branch', String(255), nullable=False),
    Column('message', Text, nullable=False),
    )
sync_query_table = Table(
    'sync_query', metadata,
    Column('key', Integer, primary_key=True),
    Column('name', String(255), index=True, unique=True, nullable=False),
    Column('updated', DateTime, index=True),
    )
file_table = Table(
    'file', metadata,
    Column('key', Integer, primary_key=True),
    Column('revision_key', Integer, ForeignKey("revision.key"), index=True),
    Column('path', Text, nullable=False, index=True),
    Column('old_path', Text, index=True),
    Column('inserted', Integer),
    Column('deleted', Integer),
    Column('status', String(1), nullable=False),
    )


class Account(object):
    def __init__(self, id, name=None, username=None, email=None):
        self.id = id
        self.name = name
        self.username = username
        self.email = email

class Project(object):
    def __init__(self, name, subscribed=False, description=''):
        self.name = name
        self.subscribed = subscribed
        self.description = description

    def delete(self):
        # This unusual delete method is to accomodate the
        # self-referential many-to-many relationship from
        # change_conflict.  With the default cascade configuration,
        # the entry from the association table is deleted regardless
        # of whether the change being deleted is change1 or change2.
        # However, when both changes are deleted at once (only likely
        # to happen when the entire project is deleted), SQLAlchemy
        # cascades through both relationships and attempts to delete
        # the entry twice.  Since we rarely delete projects, we add a
        # special case here to delete a project's changes one at a
        # time to avoid this situation.
        session = Session.object_session(self)
        for c in self.changes:
            session.delete(c)
            session.flush()
            session.expire_all()
        session.delete(self)

    def createChange(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        c = Change(*args, **kw)
        self.changes.append(c)
        session.add(c)
        session.flush()
        return c

    def createBranch(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        b = Branch(*args, **kw)
        self.branches.append(b)
        session.add(b)
        session.flush()
        return b

class Branch(object):
    def __init__(self, project, name):
        self.project_key = project.key
        self.name = name

class ProjectTopic(object):
    def __init__(self, project, topic, sequence):
        self.project_key = project.key
        self.topic_key = topic.key
        self.sequence = sequence

class Topic(object):
    def __init__(self, name, sequence):
        self.name = name
        self.sequence = sequence

    def addProject(self, project):
        session = Session.object_session(self)
        seq = max([x.sequence for x in self.project_topics] + [0])
        pt = ProjectTopic(project, self, seq+1)
        self.project_topics.append(pt)
        self.projects.append(project)
        session.add(pt)
        session.flush()

    def removeProject(self, project):
        session = Session.object_session(self)
        for pt in self.project_topics:
            if pt.project_key == project.key:
                self.project_topics.remove(pt)
                session.delete(pt)
        self.projects.remove(project)
        session.flush()

class Change(object):
    def __init__(self, project, id, owner, number, branch, change_id,
                 subject, created, updated, status, topic=None,
                 hidden=False, reviewed=False, starred=False, held=False,
                 pending_rebase=False, pending_topic=False,
                 pending_starred=False, pending_status=False,
                 pending_status_message=None, outdated=False):
        self.project_key = project.key
        self.account_key = owner.key
        self.id = id
        self.number = number
        self.branch = branch
        self.change_id = change_id
        self.topic = topic
        self.subject = subject
        self.created = created
        self.updated = updated
        self.status = status
        self.hidden = hidden
        self.reviewed = reviewed
        self.starred = starred
        self.held = held
        self.pending_rebase = pending_rebase
        self.pending_topic = pending_topic
        self.pending_starred = pending_starred
        self.pending_status = pending_status
        self.pending_status_message = pending_status_message
        self.outdated = outdated

    def getCategories(self):
        categories = set([label.category for label in self.labels])
        return sorted(categories)

    def getMaxForCategory(self, category):
        if not hasattr(self, '_approval_cache'):
            self._updateApprovalCache()
        return self._approval_cache.get(category, 0)

    def _updateApprovalCache(self):
        cat_min = {}
        cat_max = {}
        cat_value = {}
        for approval in self.approvals:
            if approval.draft:
                continue
            cur_min = cat_min.get(approval.category, 0)
            cur_max = cat_max.get(approval.category, 0)
            cur_min = min(approval.value, cur_min)
            cur_max = max(approval.value, cur_max)
            cat_min[approval.category] = cur_min
            cat_max[approval.category] = cur_max
            cur_value = cat_value.get(approval.category, 0)
            if abs(cur_min) > abs(cur_value):
                cur_value = cur_min
            if abs(cur_max) > abs(cur_value):
                cur_value = cur_max
            cat_value[approval.category] = cur_value
        self._approval_cache = cat_value

    def getMinMaxPermittedForCategory(self, category):
        if not hasattr(self, '_permitted_cache'):
            self._updatePermittedCache()
        return self._permitted_cache.get(category, (0,0))

    def _updatePermittedCache(self):
        cache = {}
        for label in self.labels:
            if label.category not in cache:
                cache[label.category] = [0, 0]
            if label.value > cache[label.category][1]:
                cache[label.category][1] = label.value
            if label.value < cache[label.category][0]:
                cache[label.category][0] = label.value
        self._permitted_cache = cache

    def createRevision(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        r = Revision(*args, **kw)
        self.revisions.append(r)
        session.add(r)
        session.flush()
        return r

    def createLabel(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        l = Label(*args, **kw)
        self.labels.append(l)
        session.add(l)
        session.flush()
        return l

    def createApproval(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        l = Approval(*args, **kw)
        self.approvals.append(l)
        session.add(l)
        session.flush()
        return l

    def createPermittedLabel(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        l = PermittedLabel(*args, **kw)
        self.permitted_labels.append(l)
        session.add(l)
        session.flush()
        return l

    @property
    def owner_name(self):
        owner_name = 'Anonymous Coward'
        if self.owner:
            if self.owner.name:
                owner_name = self.owner.name
            elif self.owner.username:
                owner_name = self.owner.username
            elif self.owner.email:
                owner_name = self.owner.email
        return owner_name

    @property
    def conflicts(self):
        return tuple(set(self.conflicts1 + self.conflicts2))

    def addConflict(self, other):
        session = Session.object_session(self)
        if other in self.conflicts1 or other in self.conflicts2:
            return
        if self in other.conflicts1 or self in other.conflicts2:
            return
        self.conflicts1.append(other)
        session.flush()
        session.expire(other, attribute_names=['conflicts2'])

    def delConflict(self, other):
        session = Session.object_session(self)
        if other in self.conflicts1:
            self.conflicts1.remove(other)
            session.flush()
            session.expire(other, attribute_names=['conflicts2'])
        if self in other.conflicts1:
            other.conflicts1.remove(self)
            session.flush()
            session.expire(self, attribute_names=['conflicts2'])

class Revision(object):
    def __init__(self, change, number, message, commit, parent,
                 fetch_auth, fetch_ref, pending_message=False,
                 can_submit=False):
        self.change_key = change.key
        self.number = number
        self.message = message
        self.commit = commit
        self.parent = parent
        self.fetch_auth = fetch_auth
        self.fetch_ref = fetch_ref
        self.pending_message = pending_message
        self.can_submit = can_submit

    def createMessage(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        m = Message(*args, **kw)
        self.messages.append(m)
        session.add(m)
        session.flush()
        return m

    def createPendingCherryPick(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        c = PendingCherryPick(*args, **kw)
        self.pending_cherry_picks.append(c)
        session.add(c)
        session.flush()
        return c

    def createFile(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        f = File(*args, **kw)
        self.files.append(f)
        session.add(f)
        session.flush()
        if hasattr(self, '_file_cache'):
            self._file_cache[f.path] = f
        return f

    def getFile(self, path):
        if not hasattr(self, '_file_cache'):
            self._file_cache = {}
            for f in self.files:
                self._file_cache[f.path] = f
        return self._file_cache.get(path, None)

    def getPendingMessage(self):
        for m in self.messages:
            if m.pending:
                return m
        return None

    def getDraftMessage(self):
        for m in self.messages:
            if m.draft:
                return m
        return None


class Message(object):
    def __init__(self, revision, id, author, created, message, draft=False, pending=False):
        self.revision_key = revision.key
        self.account_key = author.key
        self.id = id
        self.created = created
        self.message = message
        self.draft = draft
        self.pending = pending

    @property
    def author_name(self):
        author_name = 'Anonymous Coward'
        if self.author:
            if self.author.name:
                author_name = self.author.name
            elif self.author.username:
                author_name = self.author.username
            elif self.author.email:
                author_name = self.author.email
        return author_name

class Comment(object):
    def __init__(self, file, id, author, in_reply_to, created, parent, line, message, draft=False):
        self.file_key = file.key
        self.account_key = author.key
        self.id = id
        self.in_reply_to = in_reply_to
        self.created = created
        self.parent = parent
        self.line = line
        self.message = message
        self.draft = draft

class Label(object):
    def __init__(self, change, category, value, description):
        self.change_key = change.key
        self.category = category
        self.value = value
        self.description = description

class PermittedLabel(object):
    def __init__(self, change, category, value):
        self.change_key = change.key
        self.category = category
        self.value = value

class Approval(object):
    def __init__(self, change, reviewer, category, value, draft=False):
        self.change_key = change.key
        self.account_key = reviewer.key
        self.category = category
        self.value = value
        self.draft = draft

class PendingCherryPick(object):
    def __init__(self, revision, branch, message):
        self.revision_key = revision.key
        self.branch = branch
        self.message = message

class SyncQuery(object):
    def __init__(self, name):
        self.name = name

class File(object):
    STATUS_ADDED = 'A'
    STATUS_DELETED = 'D'
    STATUS_RENAMED = 'R'
    STATUS_COPIED = 'C'
    STATUS_REWRITTEN = 'W'
    STATUS_MODIFIED = 'M'

    def __init__(self, revision, path, status, old_path=None,
                 inserted=None, deleted=None):
        self.revision_key = revision.key
        self.path = path
        self.status = status
        self.old_path = old_path
        self.inserted = inserted
        self.deleted = deleted

    @property
    def display_path(self):
        if not self.old_path:
            return self.path
        pre = []
        post = []
        for start in range(min(len(self.old_path), len(self.path))):
            if self.path[start] == self.old_path[start]:
                pre.append(self.old_path[start])
            else:
                break
        pre = ''.join(pre)
        for end in range(1, min(len(self.old_path), len(self.path))-1):
            if self.path[0-end] == self.old_path[0-end]:
                post.insert(0, self.old_path[0-end])
            else:
                break
        post = ''.join(post)
        mid = '{%s => %s}' % (self.old_path[start:0-end+1], self.path[start:0-end+1])
        if pre and post:
            mid = '{%s => %s}' % (self.old_path[start:0-end+1],
                                  self.path[start:0-end+1])
            return pre + mid + post
        else:
            return '%s => %s' % (self.old_path, self.path)

    def createComment(self, *args, **kw):
        session = Session.object_session(self)
        args = [self] + list(args)
        c = Comment(*args, **kw)
        self.comments.append(c)
        session.add(c)
        session.flush()
        return c


mapper(Account, account_table)
mapper(Project, project_table, properties=dict(
    branches=relationship(Branch, backref='project',
                          order_by=branch_table.c.name,
                          cascade='all, delete-orphan'),
    changes=relationship(Change, backref='project',
                         order_by=change_table.c.number,
                         cascade='all, delete-orphan'),
    topics=relationship(Topic,
                        secondary=project_topic_table,
                        order_by=topic_table.c.name,
                        viewonly=True),
    unreviewed_changes=relationship(Change,
                                    primaryjoin=and_(project_table.c.key==change_table.c.project_key,
                                                     change_table.c.hidden==False,
                                                     change_table.c.status!='MERGED',
                                                     change_table.c.status!='ABANDONED',
                                                     change_table.c.reviewed==False),
                                    order_by=change_table.c.number,
                                ),
    open_changes=relationship(Change,
                              primaryjoin=and_(project_table.c.key==change_table.c.project_key,
                                               change_table.c.status!='MERGED',
                                               change_table.c.status!='ABANDONED'),
                              order_by=change_table.c.number,
                          ),
))
mapper(Branch, branch_table)
mapper(Topic, topic_table, properties=dict(
    projects=relationship(Project,
                          secondary=project_topic_table,
                          order_by=project_table.c.name,
                          viewonly=True),
    project_topics=relationship(ProjectTopic),
))
mapper(ProjectTopic, project_topic_table)
mapper(Change, change_table, properties=dict(
        owner=relationship(Account),
        conflicts1=relationship(Change,
                                secondary=change_conflict_table,
                                primaryjoin=change_table.c.key==change_conflict_table.c.change1_key,
                                secondaryjoin=change_table.c.key==change_conflict_table.c.change2_key,
                                ),
        conflicts2=relationship(Change,
                                secondary=change_conflict_table,
                                primaryjoin=change_table.c.key==change_conflict_table.c.change2_key,
                                secondaryjoin=change_table.c.key==change_conflict_table.c.change1_key,
                                ),
        revisions=relationship(Revision, backref='change',
                               order_by=revision_table.c.number,
                               cascade='all, delete-orphan'),
        messages=relationship(Message,
                              secondary=revision_table,
                              order_by=message_table.c.created,
                              viewonly=True),
        labels=relationship(Label, backref='change',
                            order_by=(label_table.c.category, label_table.c.value),
                            cascade='all, delete-orphan'),
        permitted_labels=relationship(PermittedLabel, backref='change',
                                      order_by=(permitted_label_table.c.category,
                                                permitted_label_table.c.value),
                                      cascade='all, delete-orphan'),
        approvals=relationship(Approval, backref='change',
                               order_by=(approval_table.c.category,
                                         approval_table.c.value),
                               cascade='all, delete-orphan'),
        draft_approvals=relationship(Approval,
                                     primaryjoin=and_(change_table.c.key==approval_table.c.change_key,
                                                      approval_table.c.draft==True),
                                     order_by=(approval_table.c.category,
                                               approval_table.c.value))
        ))
mapper(Revision, revision_table, properties=dict(
        messages=relationship(Message, backref='revision',
                              cascade='all, delete-orphan'),
        files=relationship(File, backref='revision',
                           cascade='all, delete-orphan'),
        pending_cherry_picks=relationship(PendingCherryPick, backref='revision',
                                          cascade='all, delete-orphan'),
        ))
mapper(Message, message_table, properties=dict(
        author=relationship(Account)))
mapper(File, file_table, properties=dict(
       comments=relationship(Comment, backref='file',
                             order_by=(comment_table.c.line,
                                       comment_table.c.created),
                             cascade='all, delete-orphan'),
       draft_comments=relationship(Comment,
                                   primaryjoin=and_(file_table.c.key==comment_table.c.file_key,
                                                    comment_table.c.draft==True),
                                   order_by=(comment_table.c.line,
                                             comment_table.c.created)),
       ))

mapper(Comment, comment_table, properties=dict(
        author=relationship(Account)))
mapper(Label, label_table)
mapper(PermittedLabel, permitted_label_table)
mapper(Approval, approval_table, properties=dict(
        reviewer=relationship(Account)))
mapper(PendingCherryPick, pending_cherry_pick_table)
mapper(SyncQuery, sync_query_table)

def match(expr, item):
    if item is None:
        return False
    return re.match(expr, item) is not None

@sqlalchemy.event.listens_for(sqlalchemy.engine.Engine, "connect")
def add_sqlite_match(dbapi_connection, connection_record):
    dbapi_connection.create_function("matches", 2, match)

class Database(object):
    def __init__(self, app, dburi, search):
        self.log = logging.getLogger('gertty.db')
        self.dburi = dburi
        self.search = search
        self.engine = create_engine(self.dburi)
        #metadata.create_all(self.engine)
        self.migrate(app)
        # If we want the objects returned from query() to be usable
        # outside of the session, we need to expunge them from the session,
        # and since the DatabaseSession always calls commit() on the session
        # when the context manager exits, we need to inform the session to
        # expire objects when it does so.
        self.session_factory = sessionmaker(bind=self.engine,
                                            expire_on_commit=False,
                                            autoflush=False)
        self.session = scoped_session(self.session_factory)
        self.lock = threading.Lock()

    def getSession(self):
        return DatabaseSession(self)

    def migrate(self, app):
        conn = self.engine.connect()
        context = alembic.migration.MigrationContext.configure(conn)
        current_rev = context.get_current_revision()
        self.log.debug('Current migration revision: %s' % current_rev)

        has_table = self.engine.dialect.has_table(conn, "project")

        config = alembic.config.Config()
        config.set_main_option("script_location", "gertty:alembic")
        config.set_main_option("sqlalchemy.url", self.dburi)
        config.gertty_app = app

        if current_rev is None and has_table:
            self.log.debug('Stamping database as initial revision')
            alembic.command.stamp(config, "44402069e137")
        alembic.command.upgrade(config, 'head')

class DatabaseSession(object):
    def __init__(self, database):
        self.database = database
        self.session = database.session
        self.search = database.search

    def __enter__(self):
        self.database.lock.acquire()
        self.start = time.time()
        return self

    def __exit__(self, etype, value, tb):
        if etype:
            self.session().rollback()
        else:
            self.session().commit()
        self.session().close()
        self.session = None
        end = time.time()
        self.database.log.debug("Database lock held %s seconds" % (end-self.start,))
        self.database.lock.release()

    def abort(self):
        self.session().rollback()

    def commit(self):
        self.session().commit()

    def delete(self, obj):
        self.session().delete(obj)

    def vacuum(self):
        self.session().execute("VACUUM")

    def getProjects(self, subscribed=False, unreviewed=False, topicless=False):
        """Retrieve projects.

        :param subscribed: If True limit to only subscribed projects.
        :param unreviewed: If True limit to only projects with unreviewed
            changes.
        :param topicless: If True limit to only projects without topics.
        """
        query = self.session().query(Project)
        if subscribed:
            query = query.filter_by(subscribed=subscribed)
            if unreviewed:
                query = query.filter(exists().where(Project.unreviewed_changes))
        if topicless:
            query = query.filter_by(topics=None)
        return query.order_by(Project.name).all()

    def getTopics(self):
        return self.session().query(Topic).order_by(Topic.sequence).all()

    def getProject(self, key):
        try:
            return self.session().query(Project).filter_by(key=key).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getProjectByName(self, name):
        try:
            return self.session().query(Project).filter_by(name=name).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getTopic(self, key):
        try:
            return self.session().query(Topic).filter_by(key=key).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getTopicByName(self, name):
        try:
            return self.session().query(Topic).filter_by(name=name).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getSyncQueryByName(self, name):
        try:
            return self.session().query(SyncQuery).filter_by(name=name).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return self.createSyncQuery(name)

    def getChange(self, key, lazy=True):
        query = self.session().query(Change).filter_by(key=key)
        if not lazy:
            query = query.options(joinedload(Change.revisions).joinedload(Revision.files).joinedload(File.comments))
        try:
            return query.one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getChangeByID(self, id):
        try:
            return self.session().query(Change).filter_by(id=id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getChangeIDs(self, ids):
        # Returns a set of IDs that exist in the local database matching
        # the set of supplied IDs. This is used when sync'ing the changesets
        # locally with the remote changes.
        if not ids:
            return set()
        query = self.session().query(Change.id)
        return set(ids).intersection(r[0] for r in query.all())

    def getChangesByChangeID(self, change_id):
        try:
            return self.session().query(Change).filter_by(change_id=change_id)
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getChangeByNumber(self, number):
        try:
            return self.session().query(Change).filter_by(number=number).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getPendingCherryPick(self, key):
        try:
            return self.session().query(PendingCherryPick).filter_by(key=key).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getChanges(self, query, unreviewed=False, sort_by='number'):
        self.database.log.debug("Search query: %s sort: %s" % (query, sort_by))
        q = self.session().query(Change).filter(self.search.parse(query))
        if not isinstance(sort_by, (list, tuple)):
            sort_by = [sort_by]
        if unreviewed:
            q = q.filter(change_table.c.hidden==False, change_table.c.reviewed==False)
        for s in sort_by:
            if s == 'updated':
                q = q.order_by(change_table.c.updated)
            elif s == 'last-seen':
                q = q.order_by(change_table.c.last_seen)
            elif s == 'number':
                q = q.order_by(change_table.c.number)
            elif s == 'project':
                q = q.filter(project_table.c.key == change_table.c.project_key)
                q = q.order_by(project_table.c.name)
        self.database.log.debug("Search SQL: %s" % q)
        try:
            return q.all()
        except sqlalchemy.orm.exc.NoResultFound:
            return []

    def getRevision(self, key):
        try:
            return self.session().query(Revision).filter_by(key=key).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getRevisionByCommit(self, commit):
        try:
            return self.session().query(Revision).filter_by(commit=commit).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getRevisionsByParent(self, parent):
        if isinstance(parent, six.string_types):
            parent = (parent,)
        try:
            return self.session().query(Revision).filter(Revision.parent.in_(parent)).all()
        except sqlalchemy.orm.exc.NoResultFound:
            return []

    def getRevisionByNumber(self, change, number):
        try:
            return self.session().query(Revision).filter_by(change_key=change.key, number=number).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getFile(self, key):
        try:
            return self.session().query(File).filter_by(key=key).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getComment(self, key):
        try:
            return self.session().query(Comment).filter_by(key=key).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getCommentByID(self, id):
        try:
            return self.session().query(Comment).filter_by(id=id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getMessage(self, key):
        try:
            return self.session().query(Message).filter_by(key=key).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getMessageByID(self, id):
        try:
            return self.session().query(Message).filter_by(id=id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getHeld(self):
        return self.session().query(Change).filter_by(held=True).all()

    def getOutdated(self):
        return self.session().query(Change).filter_by(outdated=True).all()

    def getPendingMessages(self):
        return self.session().query(Message).filter_by(pending=True).all()

    def getPendingTopics(self):
        return self.session().query(Change).filter_by(pending_topic=True).all()

    def getPendingRebases(self):
        return self.session().query(Change).filter_by(pending_rebase=True).all()

    def getPendingStarred(self):
        return self.session().query(Change).filter_by(pending_starred=True).all()

    def getPendingStatusChanges(self):
        return self.session().query(Change).filter_by(pending_status=True).all()

    def getPendingCherryPicks(self):
        return self.session().query(PendingCherryPick).all()

    def getPendingCommitMessages(self):
        return self.session().query(Revision).filter_by(pending_message=True).all()

    def getAccountByID(self, id, name=None, username=None, email=None):
        try:
            account = self.session().query(Account).filter_by(id=id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            account = self.createAccount(id)
        if name is not None and account.name != name:
            account.name = name
        if username is not None and account.username != username:
            account.username = username
        if email is not None and account.email != email:
            account.email = email
        return account

    def getAccountByUsername(self, username):
        try:
            return self.session().query(Account).filter_by(username=username).one()
        except sqlalchemy.orm.exc.NoResultFound:
            return None

    def getSystemAccount(self):
        return self.getAccountByID(0, 'Gerrit Code Review')

    def createProject(self, *args, **kw):
        o = Project(*args, **kw)
        self.session().add(o)
        self.session().flush()
        return o

    def createAccount(self, *args, **kw):
        a = Account(*args, **kw)
        self.session().add(a)
        self.session().flush()
        return a

    def createSyncQuery(self, *args, **kw):
        o = SyncQuery(*args, **kw)
        self.session().add(o)
        self.session().flush()
        return o

    def createTopic(self, *args, **kw):
        o = Topic(*args, **kw)
        self.session().add(o)
        self.session().flush()
        return o



"""add account table

Revision ID: 4cc9c46f9d8b
Revises: 725816dc500
Create Date: 2014-07-23 16:01:47.462597

"""

# revision identifiers, used by Alembic.
revision = '4cc9c46f9d8b'
down_revision = '725816dc500'

import warnings

from alembic import op
import sqlalchemy as sa

from gertty.dbsupport import sqlite_alter_columns, sqlite_drop_columns


def upgrade():
    sqlite_drop_columns('message', ['name'])
    sqlite_drop_columns('comment', ['name'])
    sqlite_drop_columns('approval', ['name'])
    sqlite_drop_columns('change', ['owner'])

    op.create_table('account',
    sa.Column('key', sa.Integer(), nullable=False),
    sa.Column('id', sa.Integer(), index=True, unique=True, nullable=False),
    sa.Column('name', sa.String(length=255)),
    sa.Column('username', sa.String(length=255)),
    sa.Column('email', sa.String(length=255)),
    sa.PrimaryKeyConstraint('key')
    )

    op.create_index(op.f('ix_account_name'), 'account', ['name'], unique=True)
    op.create_index(op.f('ix_account_username'), 'account', ['name'], unique=True)
    op.create_index(op.f('ix_account_email'), 'account', ['name'], unique=True)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        op.add_column('message', sa.Column('account_key', sa.Integer()))
        op.add_column('comment', sa.Column('account_key', sa.Integer()))
        op.add_column('approval', sa.Column('account_key', sa.Integer()))
        op.add_column('change', sa.Column('account_key', sa.Integer()))
    sqlite_alter_columns('message', [
            sa.Column('account_key', sa.Integer(), sa.ForeignKey('account.key'))
            ])
    sqlite_alter_columns('comment', [
            sa.Column('account_key', sa.Integer(), sa.ForeignKey('account.key'))
            ])
    sqlite_alter_columns('approval', [
            sa.Column('account_key', sa.Integer(), sa.ForeignKey('account.key'))
            ])
    sqlite_alter_columns('change', [
            sa.Column('account_key', sa.Integer(), sa.ForeignKey('account.key'))
            ])

    op.create_index(op.f('ix_message_account_key'), 'message', ['account_key'], unique=False)
    op.create_index(op.f('ix_comment_account_key'), 'comment', ['account_key'], unique=False)
    op.create_index(op.f('ix_approval_account_key'), 'approval', ['account_key'], unique=False)
    op.create_index(op.f('ix_change_account_key'), 'change', ['account_key'], unique=False)

    connection = op.get_bind()
    project = sa.sql.table('project', sa.sql.column('updated', sa.DateTime))
    connection.execute(project.update().values({'updated':None}))

    approval = sa.sql.table('approval', sa.sql.column('pending'))
    connection.execute(approval.delete().where(approval.c.pending==False))


def downgrade():
    pass



# Copyright 2014 OpenStack Foundation
# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import logging

import six
import urwid

from gertty import keymap
from gertty import mywid
from gertty import sync
from gertty.view import change as view_change
from gertty.view import mouse_scroll_decorator
import gertty.view


class ColumnInfo(object):
    def __init__(self, name, packing, value):
        self.name = name
        self.packing = packing
        self.value = value
        self.options = (packing, value)
        if packing == 'given':
            self.spacing = value + 1
        else:
            self.spacing = (value * 8) + 1


COLUMNS = [
    ColumnInfo('Number',  'given',   6),
    ColumnInfo('Subject', 'weight',  4),
    ColumnInfo('Project', 'weight',  1),
    ColumnInfo('Branch',  'weight',  1),
    ColumnInfo('Topic',   'weight',  1),
    ColumnInfo('Owner',   'weight',  1),
    ColumnInfo('Updated', 'given',  10),
    ColumnInfo('Size',    'given',   4),
]


class ThreadStack(object):
    def __init__(self):
        self.stack = []

    def push(self, change, children):
        self.stack.append([change, children])

    def pop(self):
        while self.stack:
            if self.stack[-1][1]:
                # handle children at the tip
                return self.stack[-1][1].pop(0)
            else:
                # current tip has no children, walk up
                self.stack.pop()
                continue
        return None

    def countChildren(self):
        return [len(x[1]) for x in self.stack]


class ChangeListColumns(object):
    def updateColumns(self):
        del self.columns.contents[:]
        cols = self.columns.contents
        options = self.columns.options

        for colinfo in COLUMNS:
            if colinfo.name in self.enabled_columns:
                attr = colinfo.name.lower().replace(' ', '_')
                cols.append((getattr(self, attr),
                             options(*colinfo.options)))

        for c in self.category_columns:
            cols.append(c)


class ChangeRow(urwid.Button, ChangeListColumns):
    change_focus_map = {None: 'focused',
                        'unreviewed-change': 'focused-unreviewed-change',
                        'reviewed-change': 'focused-reviewed-change',
                        'starred-change': 'focused-starred-change',
                        'held-change': 'focused-held-change',
                        'marked-change': 'focused-marked-change',
                        'positive-label': 'focused-positive-label',
                        'negative-label': 'focused-negative-label',
                        'min-label': 'focused-min-label',
                        'max-label': 'focused-max-label',


                        'added-graph': 'focused-added-graph',
                        'removed-graph': 'focused-removed-graph',
                        }

    def selectable(self):
        return True

    def __init__(self, app, change, prefix, categories,
                 enabled_columns, callback=None):
        super(ChangeRow, self).__init__('', on_press=callback, user_data=change.key)
        self.app = app
        self.change_key = change.key
        self.prefix = prefix
        self.enabled_columns = enabled_columns
        self.subject = mywid.SearchableText(u'', wrap='clip')
        self.number = mywid.SearchableText(u'')
        self.updated = mywid.SearchableText(u'')
        self.size = mywid.SearchableText(u'', align='right')
        self.project = mywid.SearchableText(u'', wrap='clip')
        self.owner = mywid.SearchableText(u'', wrap='clip')
        self.branch = mywid.SearchableText(u'', wrap='clip')
        self.topic = mywid.SearchableText(u'', wrap='clip')
        self.mark = False
        self.columns = urwid.Columns([], dividechars=1)
        self.row_style = urwid.AttrMap(self.columns, '')
        self._w = urwid.AttrMap(self.row_style, None, focus_map=self.change_focus_map)
        self.category_columns = []
        self.update(change, categories)

    def search(self, search, attribute):
        if self.subject.search(search, attribute):
            return True
        if self.number.search(search, attribute):
            return True
        if self.project.search(search, attribute):
            return True
        if self.branch.search(search, attribute):
            return True
        if self.owner.search(search, attribute):
            return True
        if self.topic.search(search, attribute):
            return True
        if self.updated.search(search, attribute):
            return True
        return False

    def _makeSize(self, added, removed):
        # Removed is a red graph on top, added is a green graph on bottom.
        #
        # The graph is 4 cells wide.  If both the red and green graphs
        # are in the cell, we set the bg to red, fg to green, and set
        # a box in the bottom half of the cell.
        #
        # If only one of the graphs is in the cell, we set a box in
        # either the top or bottom of the cell, and set the fg color
        # appropriately.  This is so that the reverse-video which
        # operates on the line when focused works as expected.

        lower_box = u'\u2584'
        upper_box = u'\u2580'
        ret = []
        # The graph is logarithmic -- one cell for each order of
        # magnitude.
        for threshold in [1, 10, 100, 1000]:
            color = []
            if (added > threshold and removed > threshold):
                ret.append(('added-removed-graph', lower_box))
            elif (added > threshold):
                ret.append(('added-graph', lower_box))
            elif (removed > threshold):
                ret.append(('removed-graph', upper_box))
            else:
                ret.append(' ')
        return ret

    def update(self, change, categories):
        if change.reviewed or change.hidden:
            style = 'reviewed-change'
        else:
            style = 'unreviewed-change'
        subject = '%s%s' % (self.prefix, change.subject)
        flag = ' '
        if change.starred:
            flag = '*'
            style = 'starred-change'
        if change.held:
            flag = '!'
            style = 'held-change'
        if self.mark:
            flag = '%'
            style = 'marked-change'
        subject = flag + subject
        self.row_style.set_attr_map({None: style})
        self.subject.set_text(subject)
        self.number.set_text(str(change.number))
        self.project.set_text(change.project.name.split('/')[-1])
        self.owner.set_text(change.owner_name)
        self.branch.set_text(change.branch or '')
        self.topic.set_text(change.topic or '')
        self.project_name = change.project.name
        self.commit_sha = change.revisions[-1].commit
        self.current_revision_key = change.revisions[-1].key
        today = self.app.time(datetime.datetime.utcnow()).date()
        updated_time = self.app.time(change.updated)
        if today == updated_time.date():
            self.updated.set_text(updated_time.strftime("%I:%M %p").upper())
        else:
            self.updated.set_text(updated_time.strftime("%Y-%m-%d"))
        total_added = 0
        total_removed = 0
        for rfile in change.revisions[-1].files:
            if rfile.status is None:
                continue
            total_added += rfile.inserted or 0
            total_removed += rfile.deleted or 0
        self.size.set_text(self._makeSize(total_added, total_removed))

        self.category_columns = []
        for category in categories:
            v = change.getMaxForCategory(category)
            cat_min, cat_max = change.getMinMaxPermittedForCategory(category)
            if v == 0:
                val = ''
            elif v > 0:
                val = '%2i' % v
                if v == cat_max:
                    val = ('max-label', val)
                else:
                    val = ('positive-label', val)
            else:
                val = '%i' % v
                if v == cat_min:
                    val = ('min-label', val)
                else:
                    val = ('negative-label', val)
            self.category_columns.append((urwid.Text(val),
                                          self.columns.options('given', 2)))
        self.updateColumns()

class ChangeListHeader(urwid.WidgetWrap, ChangeListColumns):
    def __init__(self, enabled_columns):
        self.enabled_columns = enabled_columns
        self.subject = urwid.Text(u'Subject', wrap='clip')
        self.number = urwid.Text(u'Number')
        self.updated = urwid.Text(u'Updated')
        self.size = urwid.Text(u'Size')
        self.project = urwid.Text(u'Project', wrap='clip')
        self.owner = urwid.Text(u'Owner', wrap='clip')
        self.branch = urwid.Text(u'Branch', wrap='clip')
        self.topic = urwid.Text(u'Topic', wrap='clip')
        self.columns = urwid.Columns([], dividechars=1)
        self.category_columns = []
        super(ChangeListHeader, self).__init__(self.columns)

    def update(self, categories):
        self.category_columns = []
        for category in categories:
            self.category_columns.append((urwid.Text(' %s' % category[0]),
                                          self._w.options('given', 2)))
        self.updateColumns()


@mouse_scroll_decorator.ScrollByWheel
class ChangeListView(urwid.WidgetWrap, mywid.Searchable):
    required_columns = set(['Number', 'Subject', 'Updated'])
    optional_columns = set(['Topic', 'Branch', 'Size'])

    def getCommands(self):
        if self.project_key:
            refresh_help = "Sync current project"
        else:
            refresh_help = "Sync subscribed projects"
        return [
            (keymap.TOGGLE_HELD,
             "Toggle the held flag for the currently selected change"),
            (keymap.LOCAL_CHECKOUT,
             "Checkout the most recent revision of the selected change into the local repo"),
            (keymap.TOGGLE_HIDDEN,
             "Toggle the hidden flag for the currently selected change"),
            (keymap.TOGGLE_LIST_REVIEWED,
             "Toggle whether only unreviewed or all changes are displayed"),
            (keymap.TOGGLE_REVIEWED,
             "Toggle the reviewed flag for the currently selected change"),
            (keymap.TOGGLE_STARRED,
             "Toggle the starred flag for the currently selected change"),
            (keymap.TOGGLE_MARK,
             "Toggle the process mark for the currently selected change"),
            (keymap.REFINE_CHANGE_SEARCH,
             "Refine the current search query"),
            (keymap.ABANDON_CHANGE,
             "Abandon the marked changes"),
            (keymap.EDIT_TOPIC,
             "Set the topic of the marked changes"),
            (keymap.RESTORE_CHANGE,
             "Restore the marked changes"),
            (keymap.REFRESH,
             refresh_help),
            (keymap.REVIEW,
             "Leave reviews for the marked changes"),
            (keymap.SORT_BY_NUMBER,
             "Sort changes by number"),
            (keymap.SORT_BY_UPDATED,
             "Sort changes by how recently the change was updated"),
            (keymap.SORT_BY_REVERSE,
             "Reverse the sort"),
            (keymap.LOCAL_CHERRY_PICK,
             "Cherry-pick the most recent revision of the selected change onto the local repo"),
            (keymap.INTERACTIVE_SEARCH,
             "Interactive search"),
            ]

    def help(self):
        key = self.app.config.keymap.formatKeys
        commands = self.getCommands()
        return [(c[0], key(c[0]), c[1]) for c in commands]

    def __init__(self, app, query, query_desc=None, project_key=None,
                 unreviewed=False, sort_by=None, reverse=None):
        super(ChangeListView, self).__init__(urwid.Pile([]))
        self.log = logging.getLogger('gertty.view.change_list')
        self.searchInit()
        self.app = app
        self.query = query
        self.query_desc = query_desc or query
        self.unreviewed = unreviewed
        self.change_rows = {}
        self.enabled_columns = set()
        for colinfo in COLUMNS:
            if (colinfo.name in self.required_columns or
                colinfo.name not in self.optional_columns):
                self.enabled_columns.add(colinfo.name)
        self.disabled_columns = set()
        self.listbox = urwid.ListBox(urwid.SimpleFocusListWalker([]))
        self.project_key = project_key
        if 'Project' not in self.required_columns and project_key is not None:
            self.enabled_columns.discard('Project')
            self.disabled_columns.add('Project')
        if 'Owner' not in self.required_columns and 'owner:' in query:
            # This could be or'd with something else, but probably
            # not.
            self.enabled_columns.discard('Owner')
            self.disabled_columns.add('Owner')
        self.sort_by = sort_by or app.config.change_list_options['sort-by']
        if reverse is not None:
            self.reverse = reverse
        else:
            self.reverse = app.config.change_list_options['reverse']
        self.header = ChangeListHeader(self.enabled_columns)
        self.categories = []
        self.refresh()
        self._w.contents.append((app.header, ('pack', 1)))
        self._w.contents.append((urwid.Divider(), ('pack', 1)))
        self._w.contents.append((urwid.AttrWrap(self.header, 'table-header'), ('pack', 1)))
        self._w.contents.append((self.listbox, ('weight', 1)))
        self._w.set_focus(3)

    def interested(self, event):
        if not ((self.project_key is not None and
                 isinstance(event, sync.ChangeAddedEvent) and
                 self.project_key == event.project_key)
                or
                (self.project_key is None and
                 isinstance(event, sync.ChangeAddedEvent))
                or
                (isinstance(event, sync.ChangeUpdatedEvent) and
                 event.change_key in self.change_rows.keys())):
            self.log.debug("Ignoring refresh change list due to event %s" % (event,))
            return False
        self.log.debug("Refreshing change list due to event %s" % (event,))
        return True

    def refresh(self):
        unseen_keys = set(self.change_rows.keys())
        with self.app.db.getSession() as session:
            change_list = session.getChanges(self.query, self.unreviewed,
                                             sort_by=self.sort_by)
            if self.unreviewed:
                self.title = (u'Unreviewed %d changes in %s' %
                    (len(change_list), self.query_desc))
            else:
                self.title = (u'All %d changes in %s' %
                    (len(change_list), self.query_desc))
            self.short_title = self.query_desc
            if '/' in self.short_title and ' ' not in self.short_title:
                i = self.short_title.rfind('/')
                self.short_title = self.short_title[i+1:]
            self.app.status.update(title=self.title)
            categories = set()
            for change in change_list:
                categories |= set(change.getCategories())
            self.categories = sorted(categories)
            self.chooseColumns()
            self.header.update(self.categories)
            i = 0
            if self.reverse:
                change_list.reverse()
            if self.app.config.thread_changes:
                change_list, prefixes = self._threadChanges(change_list)
            else:
                prefixes = {}
            new_rows = []
            if len(self.listbox.body):
                focus_pos = self.listbox.focus_position
                focus_row = self.listbox.body[focus_pos]
            else:
                focus_pos = 0
                focus_row = None
            for change in change_list:
                row = self.change_rows.get(change.key)
                if not row:
                    row = ChangeRow(self.app, change,
                                    prefixes.get(change.key),
                                    self.categories,
                                    self.enabled_columns,
                                    callback=self.onSelect)
                    self.listbox.body.insert(i, row)
                    self.change_rows[change.key] = row
                else:
                    row.update(change, self.categories)
                    unseen_keys.remove(change.key)
                new_rows.append(row)
                i += 1
            self.listbox.body[:] = new_rows
            if focus_row in self.listbox.body:
                pos = self.listbox.body.index(focus_row)
            else:
                pos = min(focus_pos, len(self.listbox.body)-1)
            self.listbox.body.set_focus(pos)
        for key in unseen_keys:
            row = self.change_rows[key]
            del self.change_rows[key]

    def chooseColumns(self):
        currently_enabled_columns = self.enabled_columns.copy()
        size = self.app.loop.screen.get_cols_rows()
        cols = size[0]
        for colinfo in COLUMNS:
            if (colinfo.name not in self.disabled_columns):
                cols -= colinfo.spacing
        cols -= 3 * len(self.categories)

        for colinfo in COLUMNS:
            if colinfo.name in self.optional_columns:
                if cols >= colinfo.spacing:
                    self.enabled_columns.add(colinfo.name)
                    cols -= colinfo.spacing
                else:
                    self.enabled_columns.discard(colinfo.name)
        if currently_enabled_columns != self.enabled_columns:
            self.header.updateColumns()
            for key, value in six.iteritems(self.change_rows):
                value.updateColumns()

    def getQueryString(self):
        if self.project_key is not None:
            return "project:%s %s" % (self.query_desc, self.app.config.project_change_list_query)
        return self.query

    def _threadChanges(self, changes):
        ret = []
        prefixes = {}
        stack = ThreadStack()
        children = {}
        commits = {}
        orphans = changes[:]
        for change in changes:
            for revision in change.revisions:
                commits[revision.commit] = change
        for change in changes:
            revision = change.revisions[-1]
            parent = commits.get(revision.parent, None)
            if parent:
                if parent.revisions[-1].commit != revision.parent:
                    # Our parent is an outdated revision.  This could
                    # cause a cycle, so skip.  This change will not
                    # appear in the thread, but will still appear in
                    # the list.  TODO: use color to indicate it
                    # depends on an outdated change.
                    continue
                if change in orphans:
                    orphans.remove(change)
                v = children.get(parent, [])
                v.append(change)
                children[parent] = v
        if orphans:
            change = orphans.pop(0)
        else:
            change = None
        while change:
            prefix = ''
            stack_children = stack.countChildren()
            for i, nchildren in enumerate(stack_children):
                if nchildren:
                    if i+1 == len(stack_children):
                        prefix += u'\u251c'
                    else:
                        prefix += u'\u2502'
                else:
                    if i+1 == len(stack_children):
                        prefix += u'\u2514'
                    else:
                        prefix += u' '
                if i+1 == len(stack_children):
                    prefix += u'\u2500'
                else:
                    prefix += u' '
            subject = '%s%s' % (prefix, change.subject)
            change._subject = subject
            prefixes[change.key] = prefix
            ret.append(change)
            if change in children:
                stack.push(change, children[change])
            change = stack.pop()
            if (not change) and orphans:
                change = orphans.pop(0)
        assert len(ret) == len(changes)
        return (ret, prefixes)

    def clearChangeList(self):
        for key, value in six.iteritems(self.change_rows):
            self.listbox.body.remove(value)
        self.change_rows = {}

    def getNextChangeKey(self, change_key):
        row = self.change_rows.get(change_key)
        try:
            i = self.listbox.body.index(row)
        except ValueError:
            return None
        if i+1 >= len(self.listbox.body):
            return None
        row = self.listbox.body[i+1]
        return row.change_key

    def getPrevChangeKey(self, change_key):
        row = self.change_rows.get(change_key)
        try:
            i = self.listbox.body.index(row)
        except ValueError:
            return None
        if i <= 0:
            return None
        row = self.listbox.body[i-1]
        return row.change_key

    def toggleReviewed(self, change_key):
        with self.app.db.getSession() as session:
            change = session.getChange(change_key)
            change.reviewed = not change.reviewed
            self.app.project_cache.clear(change.project)
            ret = change.reviewed
            reviewed_str = 'reviewed' if change.reviewed else 'unreviewed'
            self.log.debug("Set change %s to %s", change_key, reviewed_str)
        return ret

    def toggleStarred(self, change_key):
        with self.app.db.getSession() as session:
            change = session.getChange(change_key)
            change.starred = not change.starred
            ret = change.starred
            change.pending_starred = True
        self.app.sync.submitTask(
            sync.ChangeStarredTask(change_key, sync.HIGH_PRIORITY))
        return ret

    def toggleHeld(self, change_key):
        return self.app.toggleHeldChange(change_key)

    def toggleHidden(self, change_key):
        with self.app.db.getSession() as session:
            change = session.getChange(change_key)
            change.hidden = not change.hidden
            ret = change.hidden
            hidden_str = 'hidden' if change.hidden else 'visible'
            self.log.debug("Set change %s to %s", change_key, hidden_str)
        return ret

    def advance(self):
        pos = self.listbox.focus_position
        if pos < len(self.listbox.body)-1:
            pos += 1
            self.listbox.focus_position = pos

    def keypress(self, size, key):
        if self.searchKeypress(size, key):
            return None

        if not self.app.input_buffer:
            key = super(ChangeListView, self).keypress(size, key)
        keys = self.app.input_buffer + [key]
        commands = self.app.config.keymap.getCommands(keys)
        ret = self.handleCommands(commands)
        if ret is True:
            if keymap.FURTHER_INPUT not in commands:
                self.app.clearInputBuffer()
            return None
        return key

    def onResize(self):
        self.chooseColumns()

    def handleCommands(self, commands):
        if keymap.TOGGLE_LIST_REVIEWED in commands:
            self.unreviewed = not self.unreviewed
            self.refresh()
            return True
        if keymap.TOGGLE_REVIEWED in commands:
            if not len(self.listbox.body):
                return True
            pos = self.listbox.focus_position
            change_key = self.listbox.body[pos].change_key
            reviewed = self.toggleReviewed(change_key)
            if self.unreviewed and reviewed:
                # Here we can avoid a full refresh by just removing the particular
                # row from the change list if the view is for the unreviewed changes
                # only.
                row = self.change_rows[change_key]
                self.listbox.body.remove(row)
                del self.change_rows[change_key]
            else:
                # Just fall back on doing a full refresh if we're in a situation
                # where we're not just popping a row from the list of unreviewed
                # changes.
                self.refresh()
                self.advance()
            return True
        if keymap.TOGGLE_HIDDEN in commands:
            if not len(self.listbox.body):
                return True
            pos = self.listbox.focus_position
            change_key = self.listbox.body[pos].change_key
            hidden = self.toggleHidden(change_key)
            if hidden:
                # Here we can avoid a full refresh by just removing the particular
                # row from the change list
                row = self.change_rows[change_key]
                self.listbox.body.remove(row)
                del self.change_rows[change_key]
            else:
                # Just fall back on doing a full refresh if we're in a situation
                # where we're not just popping a row from the list of changes.
                self.refresh()
                self.advance()
            return True
        if keymap.TOGGLE_HELD in commands:
            if not len(self.listbox.body):
                return True
            pos = self.listbox.focus_position
            change_key = self.listbox.body[pos].change_key
            self.toggleHeld(change_key)
            row = self.change_rows[change_key]
            with self.app.db.getSession() as session:
                change = session.getChange(change_key)
                row.update(change, self.categories)
            self.advance()
            return True
        if keymap.TOGGLE_STARRED in commands:
            if not len(self.listbox.body):
                return True
            pos = self.listbox.focus_position
            change_key = self.listbox.body[pos].change_key
            self.toggleStarred(change_key)
            row = self.change_rows[change_key]
            with self.app.db.getSession() as session:
                change = session.getChange(change_key)
                row.update(change, self.categories)
            self.advance()
            return True
        if keymap.TOGGLE_MARK in commands:
            if not len(self.listbox.body):
                return True
            pos = self.listbox.focus_position
            change_key = self.listbox.body[pos].change_key
            row = self.change_rows[change_key]
            row.mark = not row.mark
            with self.app.db.getSession() as session:
                change = session.getChange(change_key)
                row.update(change, self.categories)
            self.advance()
            return True
        if keymap.EDIT_TOPIC in commands:
            self.editTopic()
            return True
        if keymap.REFRESH in commands:
            if self.project_key:
                self.app.sync.submitTask(
                    sync.SyncProjectTask(self.project_key, sync.HIGH_PRIORITY))
            else:
                self.app.sync.submitTask(
                    sync.SyncSubscribedProjectsTask(sync.HIGH_PRIORITY))
            self.app.status.update()
            return True
        if keymap.REVIEW in commands:
            rows = [row for row in self.change_rows.values() if row.mark]
            if not rows:
                pos = self.listbox.focus_position
                rows = [self.listbox.body[pos]]
            self.openReview(rows)
            return True
        if keymap.SORT_BY_NUMBER in commands:
            if not len(self.listbox.body):
                return True
            self.sort_by = 'number'
            self.clearChangeList()
            self.refresh()
            return True
        if keymap.SORT_BY_UPDATED in commands:
            if not len(self.listbox.body):
                return True
            self.sort_by = 'updated'
            self.clearChangeList()
            self.refresh()
            return True
        if keymap.SORT_BY_REVERSE in commands:
            if not len(self.listbox.body):
                return True
            if self.reverse:
                self.reverse = False
            else:
                self.reverse = True
            self.clearChangeList()
            self.refresh()
            return True
        if keymap.LOCAL_CHECKOUT in commands:
            if not len(self.listbox.body):
                return True
            pos = self.listbox.focus_position
            row = self.listbox.body[pos]
            self.app.localCheckoutCommit(row.project_name, row.commit_sha)
            return True
        if keymap.LOCAL_CHERRY_PICK in commands:
            if not len(self.listbox.body):
                return True
            pos = self.listbox.focus_position
            row = self.listbox.body[pos]
            self.app.localCherryPickCommit(row.project_name, row.commit_sha)
            return True
        if keymap.REFINE_CHANGE_SEARCH in commands:
            default = self.getQueryString()
            self.app.searchDialog(default)
            return True
        if keymap.ABANDON_CHANGE in commands:
            self.abandonChange()
            return True
        if keymap.RESTORE_CHANGE in commands:
            self.restoreChange()
            return True
        if keymap.INTERACTIVE_SEARCH in commands:
            self.searchStart()
            return True
        return False

    def onSelect(self, button, change_key):
        try:
            view = view_change.ChangeView(self.app, change_key)
            self.app.changeScreen(view)
        except gertty.view.DisplayError as e:
            self.app.error(str(e))

    def openReview(self, rows):
        dialog = view_change.ReviewDialog(self.app, rows[0].current_revision_key)
        urwid.connect_signal(dialog, 'save',
            lambda button: self.closeReview(dialog, rows, True, False))
        urwid.connect_signal(dialog, 'submit',
            lambda button: self.closeReview(dialog, rows, True, True))
        urwid.connect_signal(dialog, 'cancel',
            lambda button: self.closeReview(dialog, rows, False, False))
        self.app.popup(dialog,
                       relative_width=50, relative_height=75,
                       min_width=60, min_height=20)

    def closeReview(self, dialog, rows, upload, submit):
        approvals, message = dialog.getValues()
        revision_keys = [row.current_revision_key for row in rows]
        message_keys = self.app.saveReviews(revision_keys, approvals,
                                            message, upload, submit)
        if upload:
            for message_key in message_keys:
                self.app.sync.submitTask(
                    sync.UploadReviewTask(message_key, sync.HIGH_PRIORITY))
        self.refresh()
        self.app.backScreen()

    def editTopic(self):
        dialog = view_change.EditTopicDialog(self.app, '')
        urwid.connect_signal(dialog, 'save',
            lambda button: self.closeEditTopic(dialog, True))
        urwid.connect_signal(dialog, 'cancel',
            lambda button: self.closeEditTopic(dialog, False))
        self.app.popup(dialog)

    def closeEditTopic(self, dialog, save):
        if save:
            rows = [row for row in self.change_rows.values() if row.mark]
            if not rows:
                pos = self.listbox.focus_position
                rows = [self.listbox.body[pos]]
            change_keys = [row.change_key for row in rows]
            with self.app.db.getSession() as session:
                for change_key in change_keys:
                    change = session.getChange(change_key)
                    change.topic = dialog.entry.edit_text
                    change.pending_topic = True
                    self.app.sync.submitTask(
                        sync.SetTopicTask(change_key, sync.HIGH_PRIORITY))
        self.app.backScreen()
        self.refresh()

    def abandonChange(self):
        dialog = mywid.TextEditDialog(u'Abandon Change', u'Abandon message:',
                                      u'Abandon Change', u'')
        urwid.connect_signal(dialog, 'cancel', self.app.backScreen)
        urwid.connect_signal(dialog, 'save', lambda button:
                                 self.doAbandonRestoreChange(dialog, 'ABANDONED'))
        self.app.popup(dialog)

    def restoreChange(self):
        dialog = mywid.TextEditDialog(u'Restore Change', u'Restore message:',
                                      u'Restore Change', u'')
        urwid.connect_signal(dialog, 'cancel', self.app.backScreen)
        urwid.connect_signal(dialog, 'save', lambda button:
                             self.doAbandonRestoreChange(dialog, 'NEW'))
        self.app.popup(dialog)

    def doAbandonRestoreChange(self, dialog, state):
        rows = [row for row in self.change_rows.values() if row.mark]
        if not rows:
            pos = self.listbox.focus_position
            rows = [self.listbox.body[pos]]
        change_keys = [row.change_key for row in rows]
        with self.app.db.getSession() as session:
            for change_key in change_keys:
                change = session.getChange(change_key)
                change.status = state
                change.pending_status = True
                change.pending_status_message = dialog.entry.edit_text
                self.app.sync.submitTask(
                    sync.ChangeStatusTask(change_key, sync.HIGH_PRIORITY))
        self.app.backScreen()
        self.refresh()



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import re

import ply.yacc as yacc
from sqlalchemy.sql.expression import and_, or_, not_, select, func

import gertty.db
import gertty.search
from gertty.search.tokenizer import tokens  # NOQA

def age_to_delta(delta, unit):
    if unit in ['seconds', 'second', 'sec', 's']:
        pass
    elif unit in ['minutes', 'minute', 'min', 'm']:
        delta = delta * 60
    elif unit in ['hours', 'hour', 'hr', 'h']:
        delta = delta * 60 * 60
    elif unit in ['days', 'day', 'd']:
        delta = delta * 60 * 60 * 24
    elif unit in ['weeks', 'week', 'w']:
        delta = delta * 60 * 60 * 24 * 7
    elif unit in ['months', 'month', 'mon']:
        delta = delta * 60 * 60 * 24 * 30
    elif unit in ['years', 'year', 'y']:
        delta = delta * 60 * 60 * 24 * 365
    return delta

def SearchParser():
    precedence = (  # NOQA
        ('left', 'NOT', 'NEG'),
    )

    def p_terms(p):
        '''expression : list_expr
                      | paren_expr
                      | boolean_expr
                      | negative_expr
                      | term'''
        p[0] = p[1]

    def p_list_expr(p):
        '''list_expr : expression expression'''
        p[0] = and_(p[1], p[2])

    def p_paren_expr(p):
        '''paren_expr : LPAREN expression RPAREN'''
        p[0] = p[2]

    def p_boolean_expr(p):
        '''boolean_expr : expression AND expression
                        | expression OR expression'''
        if p[2].lower() == 'and':
            p[0] = and_(p[1], p[3])
        elif p[2].lower() == 'or':
            p[0] = or_(p[1], p[3])
        else:
            raise gertty.search.SearchSyntaxError("Boolean %s not recognized" % p[2])

    def p_negative_expr(p):
        '''negative_expr : NOT expression
                         | NEG expression'''
        p[0] = not_(p[2])

    def p_term(p):
        '''term : age_term
                | recentlyseen_term
                | change_term
                | owner_term
                | reviewer_term
                | commit_term
                | project_term
                | projects_term
                | project_key_term
                | branch_term
                | topic_term
                | ref_term
                | label_term
                | message_term
                | comment_term
                | has_term
                | is_term
                | status_term
                | file_term
                | limit_term
                | op_term'''
        p[0] = p[1]

    def p_string(p):
        '''string : SSTRING
                  | DSTRING
                  | USTRING'''
        p[0] = p[1]

    def p_age_term(p):
        '''age_term : OP_AGE NUMBER string'''
        now = datetime.datetime.utcnow()
        delta = p[2]
        unit = p[3]
        delta = age_to_delta(delta, unit)
        p[0] = gertty.db.change_table.c.updated < (now-datetime.timedelta(seconds=delta))

    def p_recentlyseen_term(p):
        '''recentlyseen_term : OP_RECENTLYSEEN NUMBER string'''
        # A gertty extension
        now = datetime.datetime.utcnow()
        delta = p[2]
        unit = p[3]
        delta = age_to_delta(delta, unit)
        s = select([func.datetime(func.max(gertty.db.change_table.c.last_seen), '-%s seconds' % delta)],
                   correlate=False)
        p[0] = gertty.db.change_table.c.last_seen >= s

    def p_change_term(p):
        '''change_term : OP_CHANGE CHANGE_ID
                       | OP_CHANGE NUMBER'''
        if type(p[2]) == int:
            p[0] = gertty.db.change_table.c.number == p[2]
        else:
            p[0] = gertty.db.change_table.c.change_id == p[2]

    def p_owner_term(p):
        '''owner_term : OP_OWNER string'''
        if p[2] == 'self':
            username = p.parser.username
            p[0] = gertty.db.account_table.c.username == username
        else:
            p[0] = or_(gertty.db.account_table.c.username == p[2],
                       gertty.db.account_table.c.email == p[2],
                       gertty.db.account_table.c.name == p[2])

    def p_reviewer_term(p):
        '''reviewer_term : OP_REVIEWER string
                         | OP_REVIEWER NUMBER'''
        filters = []
        filters.append(gertty.db.approval_table.c.change_key == gertty.db.change_table.c.key)
        filters.append(gertty.db.approval_table.c.account_key == gertty.db.account_table.c.key)
        try:
            number = int(p[2])
        except:
            number = None
        if number is not None:
            filters.append(gertty.db.account_table.c.id == number)
        elif p[2] == 'self':
            username = p.parser.username
            filters.append(gertty.db.account_table.c.username == username)
        else:
            filters.append(or_(gertty.db.account_table.c.username == p[2],
                               gertty.db.account_table.c.email == p[2],
                               gertty.db.account_table.c.name == p[2]))
        s = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
        p[0] = gertty.db.change_table.c.key.in_(s)

    def p_commit_term(p):
        '''commit_term : OP_COMMIT string'''
        filters = []
        filters.append(gertty.db.revision_table.c.change_key == gertty.db.change_table.c.key)
        filters.append(gertty.db.revision_table.c.commit == p[2])
        s = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
        p[0] = gertty.db.change_table.c.key.in_(s)

    def p_project_term(p):
        '''project_term : OP_PROJECT string'''
        if p[2].startswith('^'):
            p[0] = func.matches(p[2], gertty.db.project_table.c.name)
        else:
            p[0] = gertty.db.project_table.c.name == p[2]

    def p_projects_term(p):
        '''projects_term : OP_PROJECTS string'''
        p[0] = gertty.db.project_table.c.name.like('%s%%' % p[2])

    def p_project_key_term(p):
        '''project_key_term : OP_PROJECT_KEY NUMBER'''
        p[0] = gertty.db.change_table.c.project_key == p[2]

    def p_branch_term(p):
        '''branch_term : OP_BRANCH string'''
        if p[2].startswith('^'):
            p[0] = func.matches(p[2], gertty.db.change_table.c.branch)
        else:
            p[0] = gertty.db.change_table.c.branch == p[2]

    def p_topic_term(p):
        '''topic_term : OP_TOPIC string'''
        if p[2].startswith('^'):
            p[0] = func.matches(p[2], gertty.db.change_table.c.topic)
        else:
            p[0] = and_(gertty.db.change_table.c.topic.isnot(None),
                        gertty.db.change_table.c.topic == p[2])

    def p_ref_term(p):
        '''ref_term : OP_REF string'''
        if p[2].startswith('^'):
            p[0] = func.matches(p[2], 'refs/heads/'+gertty.db.change_table.c.branch)
        else:
            p[0] = gertty.db.change_table.c.branch == p[2][len('refs/heads/'):]

    label_re = re.compile(r'(?P<label>[a-zA-Z0-9_-]+([a-zA-Z]|((?<![-+])[0-9])))'
                          r'(?P<operator>[<>]?=?)(?P<value>[-+]?[0-9]+)'
                          r'($|,(user=)?(?P<user>\S+))')

    def p_label_term(p):
        '''label_term : OP_LABEL string'''
        args = label_re.match(p[2])
        label = args.group('label')
        op = args.group('operator') or '='
        value = int(args.group('value'))
        user = args.group('user')

        filters = []
        filters.append(gertty.db.approval_table.c.change_key == gertty.db.change_table.c.key)
        filters.append(gertty.db.approval_table.c.category == label)
        if op == '=':
            filters.append(gertty.db.approval_table.c.value == value)
        elif op == '>=':
            filters.append(gertty.db.approval_table.c.value >= value)
        elif op == '<=':
            filters.append(gertty.db.approval_table.c.value <= value)
        if user is not None:
            filters.append(gertty.db.approval_table.c.account_key == gertty.db.account_table.c.key)
            if user == 'self':
                filters.append(gertty.db.account_table.c.username == p.parser.username)
            else:
                filters.append(
                    or_(gertty.db.account_table.c.username == user,
                        gertty.db.account_table.c.email == user,
                        gertty.db.account_table.c.name == user))
        s = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
        p[0] = gertty.db.change_table.c.key.in_(s)

    def p_message_term(p):
        '''message_term : OP_MESSAGE string'''
        filters = []
        filters.append(gertty.db.revision_table.c.change_key == gertty.db.change_table.c.key)
        filters.append(gertty.db.revision_table.c.message.like('%%%s%%' % p[2]))
        s = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
        p[0] = gertty.db.change_table.c.key.in_(s)

    def p_comment_term(p):
        '''comment_term : OP_COMMENT string'''
        filters = []
        filters.append(gertty.db.revision_table.c.change_key == gertty.db.change_table.c.key)
        filters.append(gertty.db.revision_table.c.message == p[2])
        revision_select = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
        filters = []
        filters.append(gertty.db.revision_table.c.change_key == gertty.db.change_table.c.key)
        filters.append(gertty.db.comment_table.c.revision_key == gertty.db.revision_table.c.key)
        filters.append(gertty.db.comment_table.c.message == p[2])
        comment_select = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
        p[0] = or_(gertty.db.change_table.c.key.in_(comment_select),
                   gertty.db.change_table.c.key.in_(revision_select))

    def p_has_term(p):
        '''has_term : OP_HAS string'''
        #TODO: implement star
        if p[2] == 'draft':
            filters = []
            filters.append(gertty.db.revision_table.c.change_key == gertty.db.change_table.c.key)
            filters.append(gertty.db.message_table.c.revision_key == gertty.db.revision_table.c.key)
            filters.append(gertty.db.message_table.c.draft == True)
            s = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
            p[0] = gertty.db.change_table.c.key.in_(s)
        else:
            raise gertty.search.SearchSyntaxError('Syntax error: has:%s is not supported' % p[2])

    def p_is_term(p):
        '''is_term : OP_IS string'''
        #TODO: implement draft
        username = p.parser.username
        if p[2] == 'reviewed':
            filters = []
            filters.append(gertty.db.approval_table.c.change_key == gertty.db.change_table.c.key)
            filters.append(gertty.db.approval_table.c.value != 0)
            s = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
            p[0] = gertty.db.change_table.c.key.in_(s)
        elif p[2] == 'open':
            p[0] = gertty.db.change_table.c.status.notin_(['MERGED', 'ABANDONED'])
        elif p[2] == 'closed':
            p[0] = gertty.db.change_table.c.status.in_(['MERGED', 'ABANDONED'])
        elif p[2] == 'submitted':
            p[0] = gertty.db.change_table.c.status == 'SUBMITTED'
        elif p[2] == 'merged':
            p[0] = gertty.db.change_table.c.status == 'MERGED'
        elif p[2] == 'abandoned':
            p[0] = gertty.db.change_table.c.status == 'ABANDONED'
        elif p[2] == 'owner':
            p[0] = gertty.db.account_table.c.username == username
        elif p[2] == 'starred':
            p[0] = gertty.db.change_table.c.starred == True
        elif p[2] == 'held':
            # A gertty extension
            p[0] = gertty.db.change_table.c.held == True
        elif p[2] == 'reviewer':
            filters = []
            filters.append(gertty.db.approval_table.c.change_key == gertty.db.change_table.c.key)
            filters.append(gertty.db.approval_table.c.account_key == gertty.db.account_table.c.key)
            filters.append(gertty.db.account_table.c.username == username)
            s = select([gertty.db.change_table.c.key], correlate=False).where(and_(*filters))
            p[0] = gertty.db.change_table.c.key.in_(s)
        elif p[2] == 'watched':
            p[0] = gertty.db.project_table.c.subscribed == True
        else:
            raise gertty.search.SearchSyntaxError('Syntax error: is:%s is not supported' % p[2])

    def p_file_term(p):
        '''file_term : OP_FILE string'''
        if p[2].startswith('^'):
            p[0] = and_(or_(func.matches(p[2], gertty.db.file_table.c.path),
                            func.matches(p[2], gertty.db.file_table.c.old_path)),
                        gertty.db.file_table.c.status is not None)
        else:
            p[0] = and_(or_(gertty.db.file_table.c.path == p[2],
                            gertty.db.file_table.c.old_path == p[2]),
                        gertty.db.file_table.c.status is not None)

    def p_status_term(p):
        '''status_term : OP_STATUS string'''
        if p[2] == 'open':
            p[0] = gertty.db.change_table.c.status.notin_(['MERGED', 'ABANDONED'])
        elif p[2] == 'closed':
            p[0] = gertty.db.change_table.c.status.in_(['MERGED', 'ABANDONED'])
        else:
            p[0] = gertty.db.change_table.c.status == p[2].upper()

    def p_limit_term(p):
        '''limit_term : OP_LIMIT NUMBER'''
        # TODO: Implement this.  The sqlalchemy limit call needs to be
        # applied to the query operation and so can not be returned as
        # part of the production here.  The information would need to
        # be returned out-of-band.  In the mean time, since limits are
        # not as important in gertty, make this a no-op for now so
        # that it does not produce a syntax error.
        p[0] = (True == True)

    def p_op_term(p):
        'op_term : OP'
        raise SyntaxError()

    def p_error(p):
        if p:
            raise gertty.search.SearchSyntaxError('Syntax error at "%s" in search string "%s" (col %s)' % (
                    p.lexer.lexdata[p.lexpos:], p.lexer.lexdata, p.lexpos))
        else:
            raise gertty.search.SearchSyntaxError('Syntax error: EOF in search string')

    return yacc.yacc(debug=0, write_tables=0)



"""add conflicts table

Revision ID: 3610c2543e07
Revises: 4388de50824a
Create Date: 2016-02-05 16:43:20.047238

"""

# revision identifiers, used by Alembic.
revision = '3610c2543e07'
down_revision = '4388de50824a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('change_conflict',
    sa.Column('key', sa.Integer(), nullable=False),
    sa.Column('change1_key', sa.Integer(), sa.ForeignKey('change.key'), index=True),
    sa.Column('change2_key', sa.Integer(), sa.ForeignKey('change.key'), index=True),
    sa.PrimaryKeyConstraint('key')
    )


def downgrade():
    pass



"""add pending actions

Revision ID: 46b175bfa277
Revises: 3d429503a29a
Create Date: 2014-08-31 09:20:11.789330

"""

# revision identifiers, used by Alembic.
revision = '46b175bfa277'
down_revision = '3d429503a29a'

import warnings

from alembic import op
import sqlalchemy as sa

from gertty.dbsupport import sqlite_alter_columns


def upgrade():
    op.create_table('branch',
    sa.Column('key', sa.Integer(), nullable=False),
    sa.Column('project_key', sa.Integer(), sa.ForeignKey('project.key'), index=True, nullable=False),
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.PrimaryKeyConstraint('key')
    )
    op.create_table('pending_cherry_pick',
    sa.Column('key', sa.Integer(), nullable=False),
    sa.Column('revision_key', sa.Integer(), sa.ForeignKey('revision.key'), index=True, nullable=False),
    sa.Column('branch', sa.String(length=255), nullable=False),
    sa.Column('message', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('key')
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        op.add_column('change', sa.Column('pending_rebase', sa.Boolean()))
        op.add_column('change', sa.Column('pending_topic', sa.Boolean()))
        op.add_column('change', sa.Column('pending_status', sa.Boolean()))
        op.add_column('change', sa.Column('pending_status_message', sa.Text()))
        op.add_column('revision', sa.Column('pending_message', sa.Boolean()))

    connection = op.get_bind()
    change = sa.sql.table('change',
                          sa.sql.column('pending_rebase', sa.Boolean()),
                          sa.sql.column('pending_topic', sa.Boolean()),
                          sa.sql.column('pending_status', sa.Boolean()))
    connection.execute(change.update().values({'pending_rebase':False,
                                               'pending_topic':False,
                                               'pending_status':False}))
    revision = sa.sql.table('revision',
                            sa.sql.column('pending_message', sa.Boolean()))
    connection.execute(revision.update().values({'pending_message':False}))

    sqlite_alter_columns('change', [
        sa.Column('pending_rebase', sa.Boolean(), index=True, nullable=False),
        sa.Column('pending_topic', sa.Boolean(), index=True, nullable=False),
        sa.Column('pending_status', sa.Boolean(), index=True, nullable=False),
        ])
    sqlite_alter_columns('revision', [
        sa.Column('pending_message', sa.Boolean(), index=True, nullable=False),
        ])

def downgrade():
    pass



"""Added project updated column

Revision ID: 38104b4c1b84
Revises: 56e48a4a064a
Create Date: 2014-05-31 06:52:12.452205

"""

# revision identifiers, used by Alembic.
revision = '38104b4c1b84'
down_revision = '56e48a4a064a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('project', sa.Column('updated', sa.DateTime))

    conn = op.get_bind()
    res = conn.execute('select "key", name from project')
    for (key, name) in res.fetchall():
        q = sa.text("select max(updated) from change where project_key=:key")
        res = conn.execute(q, key=key)
        for (updated,) in res.fetchall():
            q = sa.text('update project set updated=:updated where "key"=:key')
            conn.execute(q, key=key, updated=updated)

    op.create_index(op.f('ix_project_updated'), 'project', ['updated'], unique=False)

def downgrade():
    op.drop_index(op.f('ix_project_updated'), table_name='project')
    op.drop_column('project', 'updated')



# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urwid

from gertty import keymap
from gertty.view import mouse_scroll_decorator

GLOBAL_HELP = (
    (keymap.HELP,
     "Display help"),
    (keymap.PREV_SCREEN,
     "Back to previous screen"),
    (keymap.TOP_SCREEN,
     "Back to project list"),
    (keymap.QUIT,
     "Quit Gertty"),
    (keymap.CHANGE_SEARCH,
     "Search for changes"),
    (keymap.LIST_HELD,
     "List held changes"),
    (keymap.KILL,
     "Kill to end of line (editing)"),
    (keymap.YANK,
     "Yank from kill ring (editing)"),
    (keymap.YANK_POP,
     "Replace previous yank from kill ring (editing)"),
    )

class TextButton(urwid.Button):
    def selectable(self):
        return True

    def __init__(self, text, on_press=None, user_data=None):
        super(TextButton, self).__init__('', on_press=on_press, user_data=user_data)
        self.text = urwid.Text(text)
        self._w = urwid.AttrMap(self.text, None, focus_map='focused')

class FixedButton(urwid.Button):
    def sizing(self):
        return frozenset([urwid.FIXED])

    def pack(self, size, focus=False):
        return (len(self.get_label())+4, 1)

class FixedRadioButton(urwid.RadioButton):
    def sizing(self):
        return frozenset([urwid.FIXED])

    def pack(self, size, focus=False):
        return (len(self.get_label())+4, 1)

class TableColumn(urwid.Pile):
    def pack(self, size, focus=False):
        maxcol = size[0]
        mx = max([i[0].pack((maxcol,), focus)[0] for i in self.contents])
        return (min(mx+2, maxcol), len(self.contents))

class Table(urwid.WidgetWrap):
    def __init__(self, headers=[], columns=None):
        if columns is None:
            cols = [('pack', TableColumn([('pack', w)])) for w in headers]
        else:
            cols = [('pack', TableColumn([])) for x in range(columns)]
        super(Table, self).__init__(
            urwid.Columns(cols))

    def addRow(self, cells=[]):
        for i, widget in enumerate(cells):
            self._w.contents[i][0].contents.append((widget, ('pack', None)))

class KillRing(object):
    def __init__(self):
        self.ring = []

    def kill(self, text):
        self.ring.append(text)

    def yank(self, repeat=False):
        if not self.ring:
            return None
        if repeat:
            t = self.ring.pop()
            self.ring.insert(0, t)
        return self.ring[-1]

class MyEdit(urwid.Edit):
    def __init__(self, *args, **kw):
        self.ring = kw.pop('ring', None)
        if not self.ring:
            self.ring = KillRing()
        self.last_yank = None
        super(MyEdit, self).__init__(*args, **kw)

    def keypress(self, size, key):
        (maxcol,) = size
        if self._command_map[key] == keymap.YANK:
            text = self.ring.yank()
            if text:
                self.last_yank = (self.edit_pos, self.edit_pos+len(text))
                self.insert_text(text)
            return
        if self._command_map[key] == keymap.YANK_POP:
            if not self.last_yank:
                return
            text = self.ring.yank(True)
            if text:
                self.edit_text = (self.edit_text[:self.last_yank[0]] +
                                  self.edit_text[self.last_yank[1]:])
                self.last_yank = (self.edit_pos, self.edit_pos+len(text))
                self.insert_text(text)
            return
        self.last_yank = None
        if self._command_map[key] == keymap.KILL:
            text = self.edit_text[self.edit_pos:]
            self.edit_text = self.edit_text[:self.edit_pos]
            self.ring.kill(text)
        return super(MyEdit, self).keypress(size, key)

class LineBoxTitlePropertyMixin(object):

    @property
    def title(self):
        return self._w.title_widget.text.strip()

    @title.setter
    def title(self, text):
        return self._w.set_title(text)

class SystemMessage(urwid.WidgetWrap, LineBoxTitlePropertyMixin):
    def __init__(self, message):
        w = urwid.Filler(urwid.Text(message, align='center'))
        super(SystemMessage, self).__init__(urwid.LineBox(w, u'System Message'))

@mouse_scroll_decorator.ScrollByWheel
class ButtonDialog(urwid.WidgetWrap, LineBoxTitlePropertyMixin):
    def __init__(self, title, message, entry_prompt=None,
                 entry_text='', buttons=[], ring=None):
        button_widgets = []
        for button in buttons:
            button_widgets.append(('pack', button))
        button_columns = urwid.Columns(button_widgets, dividechars=2)
        rows = []
        rows.append(urwid.Text(message))
        if entry_prompt:
            self.entry = MyEdit(entry_prompt, edit_text=entry_text, ring=ring)
            rows.append(self.entry)
        else:
            self.entry = None
        rows.append(urwid.Divider())
        rows.append(button_columns)
        listbox = urwid.ListBox(rows)
        super(ButtonDialog, self).__init__(urwid.LineBox(listbox, title))

class LineEditDialog(ButtonDialog):
    signals = ['save', 'cancel']
    def __init__(self, app, title, message, entry_prompt=None,
                 entry_text='', ring=None):
        self.app = app
        save_button = FixedButton('Save')
        cancel_button = FixedButton('Cancel')
        urwid.connect_signal(save_button, 'click',
                             lambda button:self._emit('save'))
        urwid.connect_signal(cancel_button, 'click',
                             lambda button:self._emit('cancel'))
        super(LineEditDialog, self).__init__(title, message, entry_prompt,
                                             entry_text,
                                             buttons=[save_button,
                                                      cancel_button],
                                             ring=ring)

    def keypress(self, size, key):
        if not self.app.input_buffer:
            key = super(LineEditDialog, self).keypress(size, key)
        keys = self.app.input_buffer + [key]
        commands = self.app.config.keymap.getCommands(keys)
        if keymap.ACTIVATE in commands:
            self._emit('save')
            return None
        return key

class TextEditDialog(urwid.WidgetWrap, LineBoxTitlePropertyMixin):
    signals = ['save', 'cancel']
    def __init__(self, title, prompt, button, text, ring=None):
        save_button = FixedButton(button)
        cancel_button = FixedButton('Cancel')
        urwid.connect_signal(save_button, 'click',
                             lambda button:self._emit('save'))
        urwid.connect_signal(cancel_button, 'click',
                             lambda button:self._emit('cancel'))
        button_widgets = [('pack', save_button),
                          ('pack', cancel_button)]
        button_columns = urwid.Columns(button_widgets, dividechars=2)
        rows = []
        self.entry = MyEdit(edit_text=text, multiline=True, ring=ring)
        rows.append(urwid.Text(prompt))
        rows.append(self.entry)
        rows.append(urwid.Divider())
        rows.append(button_columns)
        pile = urwid.Pile(rows)
        fill = urwid.Filler(pile, valign='top')
        super(TextEditDialog, self).__init__(urwid.LineBox(fill, title))

class MessageDialog(ButtonDialog):
    signals = ['close']
    def __init__(self, title, message):
        ok_button = FixedButton('OK')
        urwid.connect_signal(ok_button, 'click',
                             lambda button:self._emit('close'))
        super(MessageDialog, self).__init__(title, message, buttons=[ok_button])

class YesNoDialog(ButtonDialog):
    signals = ['yes', 'no']
    def __init__(self, title, message):
        yes_button = FixedButton('Yes')
        no_button = FixedButton('No')
        urwid.connect_signal(yes_button, 'click',
                             lambda button:self._emit('yes'))
        urwid.connect_signal(no_button, 'click',
                             lambda button:self._emit('no'))
        super(YesNoDialog, self).__init__(title, message, buttons=[yes_button,
                                                                   no_button])
    def keypress(self, size, key):
        r = super(YesNoDialog, self).keypress(size, key)
        if r in ('Y', 'y'):
            self._emit('yes')
            return None
        if r in ('N', 'n'):
            self._emit('no')
            return None
        return r

class SearchableText(urwid.Text):
    def set_text(self, markup):
        self._markup = markup
        super(SearchableText, self).set_text(markup)

    def search(self, search, attribute):
        if not search:
            self.set_text(self._markup)
            return
        (text, attrs) = urwid.util.decompose_tagmarkup(self._markup)
        last = 0
        found = False
        while True:
            start = text.find(search, last)
            if start < 0:
                break
            found = True
            end = start + len(search)
            i = 0
            newattrs = []
            for attr, al in attrs:
                if i + al <= start:
                    i += al
                    newattrs.append((attr, al))
                    continue
                if i >= end:
                    i += al
                    newattrs.append((attr, al))
                    continue
                before = max(start - i, 0)
                after = max(i + al - end, 0)
                if before:
                    newattrs.append((attr, before))
                newattrs.append((attribute, len(search)))
                if after:
                    newattrs.append((attr, after))
                i += al
            if i < start:
                newattrs.append((None, start-i))
                i += start-i
            if i < end:
                newattrs.append((attribute, len(search)))
            last = start + 1
            attrs = newattrs
        self._text = text
        self._attrib = attrs
        self._invalidate()
        return found

class Searchable(object):
    def searchInit(self):
        self.search = None
        self.results = []
        self.current_result = 0

    def searchValidChar(self, ch):
        return urwid.util.is_wide_char(ch, 0) or (len(ch) == 1 and ord(ch) >= 32)

    def searchKeypress(self, size, key):
        if self.search is not None:
            if self.searchValidChar(key) or key == 'backspace':
                if key == 'backspace':
                    self.search = self.search[:-1]
                else:
                    self.search += key
                self.interactiveSearch(self.search)
                return True
            else:
                commands = self.app.config.keymap.getCommands([key])
                if keymap.INTERACTIVE_SEARCH in commands:
                    self.nextSearchResult()
                    return True
                else:
                    self.app.status.update(title=self.title)
                    if not self.search:
                        self.interactiveSearch(None)
                    self.search = None
                    if key in ['enter', 'esc']:
                        return True
        return False

    def searchStart(self):
        self.search = ''
        self.app.status.update(title=("Search: "))

    def interactiveSearch(self, search):
        if search is not None:
            self.app.status.update(title=("Search: " + search))
        self.results = []
        self.current_result = 0
        for i, line in enumerate(self.listbox.body):
            if hasattr(line, 'search'):
                if line.search(search, 'search-result'):
                    self.results.append(i)

    def nextSearchResult(self):
        if not self.results:
            return
        dest = self.results[self.current_result]
        self.listbox.set_focus(dest)
        self.listbox._invalidate()
        self.current_result += 1
        if self.current_result >= len(self.results):
            self.current_result = 0

class HyperText(urwid.Text):
    _selectable = True

    def __init__(self, markup, align=urwid.LEFT, wrap=urwid.SPACE, layout=None):
        self._mouse_press_item = None
        self.selectable_items = []
        self.focused_index = None
        self.last_focused_index = 0
        super(HyperText, self).__init__(markup, align, wrap, layout)

    def focusFirstItem(self):
        if len(self.selectable_items) == 0:
            return False
        self.focusItem(0)
        return True

    def focusLastItem(self):
        if len(self.selectable_items) == 0:
            return False
        self.focusItem(len(self.selectable_items)-1)
        return True

    def focusPreviousItem(self):
        if len(self.selectable_items) == 0:
            return False
        if self.focused_index is None:
            self.focusItem(self.last_focused_index)
        item = max(0, self.focused_index-1)
        if item != self.focused_index:
            self.focusItem(item)
            return True
        return False

    def focusNextItem(self):
        if len(self.selectable_items) == 0:
            return False
        if self.focused_index is None:
            self.focusItem(self.last_focused_index)
        item = min(len(self.selectable_items)-1, self.focused_index+1)
        if item != self.focused_index:
            self.focusItem(item)
            return True
        return False

    def focusItem(self, item):
        self.last_focused_index = self.focused_index
        self.focused_index = item
        self.set_text(self._markup)
        self._invalidate()

    def select(self):
        if self.focused_index is not None:
            self.selectable_items[self.focused_index][0].select()

    def keypress(self, size, key):
        if self._command_map[key] == urwid.CURSOR_UP:
            if self.focusPreviousItem():
                return False
            return key
        elif self._command_map[key] == urwid.CURSOR_DOWN:
            if self.focusNextItem():
                return False
            return key
        elif self._command_map[key] == urwid.ACTIVATE:
            self.select()
            return False
        return key

    def getPosAtCoords(self, maxcol, col, row):
        trans = self.get_line_translation(maxcol)
        colpos = 0
        line = None
        try:
            line = trans[row]
        except IndexError:
            return None
        for t in line:
            if len(t) == 2:
                width, pos = t
                if colpos <= col < colpos + width:
                    return pos
            else:
                width, start, end = t
                if colpos <= col < colpos + width:
                    return start + (col - colpos)
            colpos += width
        return None

    def getItemAtCoords(self, maxcol, col, row):
        pos = self.getPosAtCoords(maxcol, col, row)
        index = 0
        for item, start, end in self.selectable_items:
            if start <= pos <= end:
                return index
            index += 1
        return None

    def mouse_event(self, size, event, button, col, row, focus):
        if ((button not in [0, 1]) or
            (event not in ['mouse press', 'mouse release'])):
            return False
        item = self.getItemAtCoords(size[0], col, row)
        if item is None:
            if self.focused_index is None:
                self.focusFirstItem()
            return False
        if event == 'mouse press':
            self.focusItem(item)
            self._mouse_press_item = item
        if event == 'mouse release':
            if self._mouse_press_item == item:
                self.select()
            self._mouse_press_item = None
        return True

    def processLinks(self, markup, data=None):
        if data is None:
            data = dict(pos=0)
        if isinstance(markup, list):
            return [self.processLinks(i, data) for i in markup]
        if isinstance(markup, tuple):
            return (markup[0], self.processLinks(markup[1], data))
        if isinstance(markup, Link):
            self.selectable_items.append((markup, data['pos'], data['pos']+len(markup.text)))
            data['pos'] += len(markup.text)
            focused = len(self.selectable_items)-1 == self.focused_index
            link_attr = markup.getAttr(focused)
            if link_attr:
                return (link_attr, markup.text)
            else:
                return markup.text
        data['pos'] += len(markup)
        return markup

    def set_text(self, markup):
        self._markup = markup
        self.selectable_items = []
        super(HyperText, self).set_text(self.processLinks(markup))

    def move_cursor_to_coords(self, size, col, row):
        if self.focused_index is None:
            if row:
                self.focusLastItem()
            else:
                self.focusFirstItem()
        return True

    def render(self, size, focus=False):
        if (not focus) and (self.focused_index is not None):
            self.focusItem(None)
        return super(HyperText, self).render(size, focus)

class Link(urwid.Widget):
    signals = ['selected']

    def __init__(self, text, attr=None, focused_attr=None):
        self.text = text
        self.attr = attr
        self.focused_attr = focused_attr

    def select(self):
        self._emit('selected')

    def getAttr(self, focus):
        if focus:
            return self.focused_attr
        return self.attr

# A workaround for the issue fixed in
# https://github.com/wardi/urwid/pull/74
# included here until thi fix is released
class MyGridFlow(urwid.GridFlow):
    def generate_display_widget(self, size):
        p = super(MyGridFlow, self).generate_display_widget(size)
        for item in p.contents:
            if isinstance(item[0], urwid.Padding):
                c = item[0].original_widget
                if isinstance(c, urwid.Columns):
                    if c.focus_position == 0 and not c.contents[0][0].selectable():
                        for i, w in enumerate(c.contents):
                            if w[0].selectable():
                                c.focus_position = i
                                break
        return p



# Copyright 2014 OpenStack Foundation
# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urwid

from gertty import gitrepo
from gertty import mywid
from gertty.view.diff import BaseDiffCommentEdit, BaseDiffComment, BaseDiffLine
from gertty.view.diff import BaseFileHeader, BaseFileReminder, BaseDiffView

LN_COL_WIDTH = 5

class UnifiedDiffCommentEdit(BaseDiffCommentEdit):
    def __init__(self, app, context, oldnew, key=None, comment=u''):
        super(UnifiedDiffCommentEdit, self).__init__([])
        self.context = context
        self.oldnew = oldnew
        # If we save a comment, the resulting key will be stored here
        self.key = key
        self.comment = mywid.MyEdit(edit_text=comment, multiline=True,
                                    ring=app.ring)
        self.contents.append((urwid.Text(u''), ('given', 8, False)))
        self.contents.append((urwid.AttrMap(self.comment, 'draft-comment'),
                              ('weight', 1, False)))
        self.focus_position = 1

class UnifiedDiffComment(BaseDiffComment):
    def __init__(self, context, oldnew, comment):
        super(UnifiedDiffComment, self).__init__([])
        self.context = context
        text = urwid.AttrMap(urwid.Text(comment), 'comment')
        self.contents.append((urwid.Text(u''), ('given', 8, False)))
        self.contents.append((text, ('weight', 1, False)))

class UnifiedDiffLine(BaseDiffLine):
    def __init__(self, app, context, oldnew, old, new, callback=None):
        super(UnifiedDiffLine, self).__init__('', on_press=callback)
        self.context = context
        self.oldnew = oldnew
        (old_ln, old_action, old_line) = old
        (new_ln, new_action, new_line) = new
        if old_ln is None:
            old_ln = ''
        else:
            old_ln = '%*i' % (LN_COL_WIDTH-1, old_ln)
        if new_ln is None:
            new_ln = ''
        else:
            new_ln = '%*i' % (LN_COL_WIDTH-1, new_ln)
        old_ln_col = urwid.Text(('line-number', old_ln))
        old_ln_col.set_wrap_mode('clip')
        new_ln_col = urwid.Text(('line-number', new_ln))
        new_ln_col.set_wrap_mode('clip')
        if oldnew == gitrepo.OLD:
            action = old_action
            line = old_line
            columns = [(LN_COL_WIDTH, old_ln_col), (LN_COL_WIDTH, urwid.Text(u''))]
        elif oldnew == gitrepo.NEW:
            action = new_action
            line = new_line
            columns = [(LN_COL_WIDTH, urwid.Text(u'')), (LN_COL_WIDTH, new_ln_col)]
        if new_action == ' ':
            columns = [(LN_COL_WIDTH, old_ln_col), (LN_COL_WIDTH, new_ln_col)]
        line_col = mywid.SearchableText(line)
        self.text_widget = line_col
        if action == '':
            line_col = urwid.AttrMap(line_col, 'nonexistent')
        columns += [line_col]
        col = urwid.Columns(columns)
        map = {None: 'focused',
               'added-line': 'focused-added-line',
               'added-word': 'focused-added-word',
               'removed-line': 'focused-removed-line',
               'removed-word': 'focused-removed-word',
               'nonexistent': 'focused-nonexistent',
               'line-number': 'focused-line-number',
               }
        self._w = urwid.AttrMap(col, None, focus_map=map)

    def search(self, search, attribute):
        return self.text_widget.search(search, attribute)

class UnifiedFileHeader(BaseFileHeader):
    def __init__(self, app, context, oldnew, old, new, callback=None):
        super(UnifiedFileHeader, self).__init__('', on_press=callback)
        self.context = context
        self.oldnew = oldnew
        if oldnew == gitrepo.OLD:
            col = urwid.Columns([
                    urwid.Text(('filename', old))])
        elif oldnew == gitrepo.NEW:
            col = urwid.Columns([
                    (LN_COL_WIDTH, urwid.Text(u'')),
                    urwid.Text(('filename', new))])
        map = {None: 'focused-filename',
               'filename': 'focused-filename'}
        self._w = urwid.AttrMap(col, None, focus_map=map)

class UnifiedFileReminder(BaseFileReminder):
    def __init__(self):
        self.old_text = urwid.Text(('filename', ''))
        self.new_text = urwid.Text(('filename', ''))
        self.col = urwid.Columns([('pack', self.old_text),
                                  ('pack', self.new_text),
                                  urwid.Text(u'')], dividechars=2)
        super(UnifiedFileReminder, self).__init__(self.col)

    def set(self, old, new):
        self.old_text.set_text(('filename', old))
        self.new_text.set_text(('filename', new))
        self.col._invalidate()

class UnifiedDiffView(BaseDiffView):
    def makeLines(self, diff, lines_to_add, comment_lists):
        lines = []
        old_cache = []
        new_cache = []
        for old, new in lines_to_add:
            context = self.makeContext(diff, old[0], new[0])
            if context.old_ln is not None:
                old_cache.append(UnifiedDiffLine(self.app, context, gitrepo.OLD, old, new,
                                                 callback=self.onSelect))
            else:
                lines.extend(old_cache)
                lines.extend(new_cache)
                old_cache = []
                new_cache = []
            # see if there are any comments for this line
            key = 'old-%s-%s' % (old[0], diff.oldname)
            old_list = comment_lists.pop(key, [])
            while old_list:
                (old_comment_key, old_comment) = old_list.pop(0)
                old_cache.append(UnifiedDiffComment(context, gitrepo.OLD, old_comment))
            # see if there are any draft comments for this line
            key = 'olddraft-%s-%s' % (old[0], diff.oldname)
            old_list = comment_lists.pop(key, [])
            while old_list:
                (old_comment_key, old_comment) = old_list.pop(0)
                old_cache.append(UnifiedDiffCommentEdit(self.app,
                                                    context,
                                                    gitrepo.OLD,
                                                    old_comment_key,
                                                    old_comment))
            # new line
            if context.new_ln is not None and new[1] != ' ':
                if old_cache:
                    new_cache.append(UnifiedDiffLine(self.app, context, gitrepo.NEW, old, new,
                                                     callback=self.onSelect))
                else:
                    lines.append(UnifiedDiffLine(self.app, context, gitrepo.NEW, old, new,
                                                 callback=self.onSelect))
            # see if there are any comments for this line
            key = 'new-%s-%s' % (new[0], diff.newname)
            new_list = comment_lists.pop(key, [])
            while new_list:
                (new_comment_key, new_comment) = new_list.pop(0)
                if old_cache:
                    new_cache.append(UnifiedDiffComment(context, gitrepo.NEW, new_comment))
                else:
                    lines.append(UnifiedDiffComment(context, gitrepo.NEW, new_comment))
            # see if there are any draft comments for this line
            key = 'newdraft-%s-%s' % (new[0], diff.newname)
            new_list = comment_lists.pop(key, [])
            while new_list:
                (new_comment_key, new_comment) = new_list.pop(0)
                if old_cache:
                    new_cache.append(UnifiedDiffCommentEdit(self.app,
                                                            context,
                                                            gitrepo.NEW,
                                                            new_comment_key,
                                                            new_comment))
                else:
                    lines.append(UnifiedDiffCommentEdit(self.app,
                                                        context,
                                                        gitrepo.NEW,
                                                        new_comment_key,
                                                        new_comment))
        else:
            if old_cache:
                lines.extend(old_cache)
            if new_cache:
                lines.extend(new_cache)
        return lines

    def makeFileReminder(self):
        return UnifiedFileReminder()

    def makeFileHeader(self, diff, comment_lists):
        context = self.makeContext(diff, None, None, header=True)
        lines = []
        lines.append(UnifiedFileHeader(self.app, context, gitrepo.OLD,
                                       diff.oldname, diff.newname,
                                       callback=self.onSelect))
        # see if there are any comments for this file
        key = 'old-None-%s' % (diff.oldname,)
        old_list = comment_lists.pop(key, [])
        while old_list:
            (old_comment_key, old_comment) = old_list.pop(0)
            lines.append(UnifiedDiffComment(context, gitrepo.OLD, old_comment))
        # see if there are any draft comments for this file
        key = 'olddraft-None-%s' % (diff.oldname,)
        old_list = comment_lists.pop(key, [])
        while old_list:
            (old_comment_key, old_comment) = old_list.pop(0)
            lines.append(UnifiedDiffCommentEdit(self.app,
                                                context,
                                                gitrepo.OLD,
                                                old_comment_key,
                                                old_comment))
        # new line
        lines.append(UnifiedFileHeader(self.app, context, gitrepo.NEW,
                                       diff.oldname, diff.newname,
                                       callback=self.onSelect))

        # see if there are any comments for this file
        key = 'new-None-%s' % (diff.newname,)
        new_list = comment_lists.pop(key, [])
        while new_list:
            (new_comment_key, new_comment) = new_list.pop(0)
            lines.append(UnifiedDiffComment(context, gitrepo.NEW, new_comment))
        # see if there are any draft comments for this file
        key = 'newdraft-None-%s' % (diff.newname,)
        new_list = comment_lists.pop(key, [])
        while new_list:
            (new_comment_key, new_comment) = new_list.pop(0)
            lines.append(UnifiedDiffCommentEdit(self.app,
                                                context,
                                                gitrepo.NEW,
                                                new_comment_key,
                                                new_comment))
        return lines

    def makeCommentEdit(self, edit):
        return UnifiedDiffCommentEdit(self.app,
                                      edit.context,
                                      edit.oldnew)

    def cleanupEdit(self, edit):
        if edit.key:
            self.deleteComment(edit.key)
            edit.key = None
        comment = edit.comment.edit_text.strip()
        if comment:
            new = False
            if edit.oldnew == gitrepo.NEW:
                new = True
            edit.key = self.saveComment(
                edit.context, comment, new=new)
        else:
            self.listbox.body.remove(edit)
